from helical.utils.mapping import map_gene_symbols_to_ensembl_ids
from helical.utils.mapping import map_ensembl_ids_to_gene_symbols
from helical.utils.mapping import convert_list_ensembl_ids_to_gene_symbols, convert_list_gene_symbols_to_ensembl_ids
import anndata as ad
import pytest

adata = ad.read_h5ad("ci/tests/data/cell_type_sample.h5ad")


def test_map_gene_symbols_to_ensembl_ids():
    """
    Test if the gene symbols are mapped to Ensembl IDs correctly.
    CD99 should be mapped to ENSG00000002586.
    """
    adata.var["gene_names"] = ["CD99"] * adata.var.shape[0]
    map_gene_symbols_to_ensembl_ids(adata, gene_names="gene_names", species='hsapiens')
    assert all(adata.var["ensembl_id"] == ["ENSG00000002586"] * adata.var.shape[0])


def test_map_ensembl_ids_to_gene_symbols():
    """
    Test if the Ensembl IDs are mapped to gene symbols correctly.
    ENSG00000002330 should be mapped to BAD.
    """
    adata.var["ensembl_id"] = ["ENSG00000002330"] * adata.var.shape[0]
    map_ensembl_ids_to_gene_symbols(adata, ensembl_id_key="ensembl_id", species='hsapiens')
    assert all(adata.var["gene_names"] == ["BAD"] * adata.var.shape[0])


@pytest.mark.skip(
    reason="This test may take a long time to run because the dataset of macaque needs to be downloaded."
)
def test_map_gene_symbols_to_ensembl_ids_macaque():
    """
    Test if the gene symbols are mapped to Ensembl IDs correctly for macaque.
    CD99 should be mapped to ENSMFAG00000000608.
    Note, this test may be long the first time it is being run because the database for macaque needs to be downloaded.
    """
    adata.var["gene_names"] = ["CD99"] * adata.var.shape[0]
    map_gene_symbols_to_ensembl_ids(adata, gene_names="gene_names", species='mfascicularis')
    assert all(adata.var["ensembl_id"] == ["ENSMFAG00000000608"] * adata.var.shape[0])


def test_convert_list_ensembl_ids_to_gene_symbols():
    ensembl_ids = ["ENSG00000139618", "ENSG00000139620"]
    gene_symbols = convert_list_ensembl_ids_to_gene_symbols(ensembl_ids, species='hsapiens')
    assert gene_symbols == ["BRCA2", "KANSL2"]


def test_convert_list_gene_symbols_to_ensembl_ids():
    gene_symbols = ["BRCA2", "KANSL2"]
    ensembl_ids = convert_list_gene_symbols_to_ensembl_ids(gene_symbols, species='hsapiens')
    assert ensembl_ids == ["ENSG00000139618", "ENSG00000139620"]



from helical.models.base_models import HelicalRNAModel
import logging
import numpy as np
from anndata import AnnData
from helical.utils.downloader import Downloader
from helical.models.genept.genept_config import GenePTConfig
from helical.utils.mapping import map_ensembl_ids_to_gene_symbols
import scanpy as sc
import torch
import json

LOGGER = logging.getLogger(__name__)


class GenePT(HelicalRNAModel):
    """GenePT Model.

    ```

    Parameters
    ----------
    configurer : GenePTConfig, optional, default = default_configurer
        The model configuration

    Notes
    -----


    """

    default_configurer = GenePTConfig()

    def __init__(self, configurer: GenePTConfig = default_configurer):
        super().__init__()
        self.configurer = configurer
        self.config = configurer.config

        downloader = Downloader()
        for file in self.config["list_of_files_to_download"]:
            downloader.download_via_name(file)

        with open(self.config["embeddings_path"], "r") as f:
            self.embeddings = json.load(f)

        LOGGER.info("GenePT initialized successfully.")

    def process_data(
        self,
        adata: AnnData,
        gene_names: str = "index",
        use_raw_counts: bool = True,
    ) -> AnnData:
        """
        Processes the data for the GenePT model.

        Parameters
        ----------
        adata : AnnData
            The AnnData object containing the data to be processed. GenePT uses Ensembl IDs to identify genes
            and currently supports only human genes. If the AnnData object already has an 'ensembl_id' column,
            the mapping step can be skipped.
        gene_names : str, optional, default="index"
            The column in `adata.var` that contains the gene names. If set to a value other than "ensembl_id",
            the gene symbols in that column will be mapped to Ensembl IDs using the 'pyensembl' package,
            which retrieves mappings from the Ensembl FTP server and loads them into a local database.
            - If set to "index", the index of the AnnData object will be used and mapped to Ensembl IDs.
            - If set to "ensembl_id", no mapping will occur.
            Special case:
                If the index of `adata` already contains Ensembl IDs, setting this to "index" will result in
                invalid mappings. In such cases, create a new column containing Ensembl IDs and pass "ensembl_id"
                as the value of `gene_names`.
        use_raw_counts : bool, optional, default=True
            Determines whether raw counts should be used.

        Returns
        -------
        Dataset
            The tokenized dataset in the form of a Huggingface Dataset object.
        """
        LOGGER.info("Processing data for GenePT.")
        self.ensure_rna_data_validity(adata, gene_names, use_raw_counts)

        # map gene symbols to ensemble ids if provided
        if gene_names == "ensembl_id":
            if (adata.var[gene_names].str.startswith("ENS").all()) or (
                adata.var[gene_names].str.startswith("None").any()
            ):
                message = (
                    "It seems an anndata with 'ensemble ids' and/or 'None' was passed. "
                    "Please set gene_names='ensembl_id' and remove 'None's to skip mapping."
                )
                LOGGER.info(message)
                raise ValueError(message)
            adata = map_ensembl_ids_to_gene_symbols(adata, gene_names)

        sc.pp.highly_variable_genes(adata, flavor="seurat_v3")
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)

        genes_names = adata.var_names[adata.var["highly_variable"]].tolist()
        adata = adata[:, genes_names]

        LOGGER.info("Successfully processed the data for GenePT.")
        return adata

    def get_text_embeddings(self, dataset: AnnData) -> np.array:
        """Gets the gene embeddings from the GenePT model

        Parameters
        ----------
        dataset : AnnData
            The tokenized dataset containing the processed data

        Returns
        -------
        np.array
            The gene embeddings in the form of a numpy array
        """
        # Generate a response
        raw_embeddings = dataset.var_names
        weights = []
        gene_list = []
        count_missed = 0

        for emb in raw_embeddings:
            gene = self.embeddings.get(emb.upper(), None)
            if gene is not None:
                weights.append(gene["embeddings"])
                gene_list.append(emb)
            else:
                count_missed += 1

        LOGGER.info(f"Couln't find {count_missed} genes in embeddings")

        weights = torch.Tensor(weights)
        embeddings = torch.matmul(
            torch.Tensor(dataset[:, gene_list].X.toarray()), weights
        )
        return embeddings

    def get_embeddings(self, dataset: AnnData) -> torch.Tensor:
        """Gets the gene embeddings from the GenePT model

        Parameters
        ----------
        dataset : Dataset
            The tokenized dataset containing the processed data

        Returns
        -------
        np.array
            The gene embeddings in the form of a numpy array
        """
        LOGGER.info(f"Inference started:")
        # Generate a response
        embeddings = self.get_text_embeddings(dataset)
        embeddings = embeddings / (np.linalg.norm(embeddings, axis=1)).reshape(-1, 1)
        return embeddings



import json
import logging
import os
import numpy as np
import torch

from helical.models.transcriptformer.utils.utils import load_embeddings

SPECIAL_TOKENS = ["unknown", "[START]", "[END]", "[RD]", "[CELL]", "[PAD]", "[MASK]"]


def construct_gene_embeddings(
    pretrained_embedding_paths: list,
    special_tokens: list,
    random_seed: int = 42,  # Add random seed parameter with default value
) -> tuple[dict, np.ndarray]:
    """Construct gene embeddings from pretrained embeddings.

    Args:
        pretrained_embedding_paths (list): List of paths to pretrained embeddings.
        special_tokens (list): List of special tokens.
        random_seed (int, optional): Random seed for special token embeddings. Defaults to 42.

    Returns
    -------
        tuple: A tuple containing:
            - gene_vocab_dict (dict): A dictionary mapping gene names to indices.
            - embeddings (np.ndarray):
    """
    gene_names = []
    embeddings = []
    for file in pretrained_embedding_paths:
        gene_emb = load_embeddings(file)
        gene_names += list(gene_emb.keys())
        embeddings += list(gene_emb.values())

    # Check for duplicates in gene names
    seen_genes = {}
    deduped_names = []
    deduped_embeddings = []
    for i, gene in enumerate(gene_names):
        if gene in seen_genes:
            logging.warning(f"Duplicate gene name found: {gene}, removing duplicate")
        else:
            deduped_names.append(gene)
            deduped_embeddings.append(embeddings[i])
            seen_genes[gene] = True

    gene_vocab_dict = build_gene_vocab_from_list(deduped_names, special_tokens)
    emb_dim = embeddings[0].shape[0]

    # Set random seed before generating special token embeddings
    rng = np.random.RandomState(random_seed)
    special_tokens_emb = rng.rand(len(special_tokens), emb_dim)

    return gene_vocab_dict, np.concatenate(
        [special_tokens_emb, np.array(deduped_embeddings)], axis=0
    )


def load_vocabs_and_embeddings(cfg):
    if cfg.model.data_config.aux_cols is not None:
        aux_cols = cfg.model.data_config.aux_cols.split(",")
    else:
        aux_cols = None

    # Load vocabularies
    aux_vocab = (
        open_vocabs(cfg.model.data_config.aux_vocab_path, aux_cols)
        if aux_cols is not None
        else None
    )

    # Load embeddings
    logging.info(
        f"Loading ESM2 mappings from {cfg.model.data_config.esm2_mappings_path}"
    )
    emb_files = []
    for file in cfg.model.data_config.esm2_mappings:
        emb_files.append(os.path.join(cfg.model.data_config.esm2_mappings_path, file))

    gene_vocab, emb_matrix = construct_gene_embeddings(
        emb_files,
        cfg.model.data_config.special_tokens,
    )

    emb_matrix = torch.tensor(emb_matrix)
    assert emb_matrix.shape[0] == len(
        gene_vocab
    ), f"Embeddings matrix has wrong shape, {emb_matrix.shape[0]} != {len(gene_vocab)}"

    return (gene_vocab, aux_vocab), emb_matrix


def build_gene_vocab_from_list(gene_names, special_tokens=None):
    logging.info("Building gene vocabulary")
    if special_tokens is None:
        special_tokens = SPECIAL_TOKENS
    gene_vocab = {token: idx for idx, token in enumerate(special_tokens)}
    gene_ids = range(len(gene_vocab), len(gene_names) + len(gene_vocab))
    gene_vocab.update(dict(zip(gene_names, gene_ids, strict=False)))
    return gene_vocab


def open_vocabs(path, cols_to_load=None, verbose=True):
    if isinstance(cols_to_load, str):
        cols_to_load = [cols_to_load]
    vocabs = {}
    for file_name in os.listdir(path):
        if file_name.endswith("_vocab.json"):
            variable_name = file_name.split("_vocab.json")[0]
            if cols_to_load is not None and variable_name not in cols_to_load:
                continue
            if verbose:
                logging.info(
                    f"Loading vocabulary file: {os.path.join(path, variable_name)}"
                )
            with open(os.path.join(path, file_name)) as f:
                vocabs[variable_name] = json.load(f)
    if len(vocabs) == 0:
        raise ValueError("No vocabularies found in the specified path")
    elif cols_to_load == ["gene"]:
        return list(vocabs.values()).pop()
    return vocabs



from helical.models.hyena_dna import HyenaDNAConfig, HyenaDNA
import pytest
import numpy as np
from pandas import DataFrame


@pytest.mark.parametrize(
    "model_name, d_model, d_inner",
    [
        ("hyenadna-tiny-1k-seqlen", 128, 512),
        ("hyenadna-tiny-1k-seqlen-d256", 256, 1024),
    ],
)
def test_hyena_dna__valid_model_names(model_name, d_model, d_inner):
    """
    Test case for the HyenaDNA class initialization.

    Args:
        model_name (str): The name of the model.
        d_model (int): The dimensionality of the model.
        d_inner (int): The dimensionality of the inner layers.
    """
    configurer = HyenaDNAConfig(model_name=model_name)
    assert configurer.config["model_path"].name == f"{model_name}.ckpt"
    assert configurer.config["d_model"] == d_model
    assert configurer.config["d_inner"] == d_inner


@pytest.mark.parametrize("model_name", ["wrong_name"])
def test_hyena_dna__invalid_model_names(model_name):
    """
    Test case when an invalid model name is provided.
    Verifies that a ValueError is raised when an invalid model name is passed to the HyenaDNAConfig constructor.

    Parameters:
    - model_name (str): The invalid model name.

    Raises:
    - ValueError: If the model name is invalid.
    """
    with pytest.raises(ValueError):
        HyenaDNAConfig(model_name=model_name)


@pytest.mark.parametrize(
    "input_sequence, expected_output",
    [
        # Valid DNA sequences
        ("", [0, 1]),
        ("A", [0, 7, 1]),
        ("CC", [0, 8, 8, 1]),
        ("TTTT", [0, 10, 10, 10, 10, 1]),
        ("ACGTN", [0, 7, 8, 9, 10, 11, 1]),
        ("ACGT" * 256, [0] + [7, 8, 9, 10] * 256 + [1]),
    ],
)
def test_hyena_dna_process_data(input_sequence, expected_output):
    model = HyenaDNA()
    output = model.process_data([input_sequence])
    expected = np.array(expected_output)
    assert np.all(np.equal(np.array(output["input_ids"][0]), expected))


@pytest.mark.parametrize(
    "input_sequence, expected_output",
    [
        (
            ["A", "CC", "TTTT", "ACGTN", "ACGT"],
            [
                [4, 4, 4, 4, 0, 7, 1],
                [4, 4, 4, 0, 8, 8, 1],
                [4, 0, 10, 10, 10, 10, 1],
                [0, 7, 8, 9, 10, 11, 1],
                [4, 0, 7, 8, 9, 10, 1],
            ],
        )
    ],
)
def test_hyena_dna_process_data_variable_length_sequences(
    input_sequence, expected_output
):
    model = HyenaDNA()
    dataset = model.process_data(input_sequence)
    assert np.all(np.equal(np.array(expected_output), np.array(dataset["input_ids"])))


@pytest.mark.parametrize(
    "data, raise_exception",
    [
        (DataFrame({"Sequence": ["ACGTN", "ACGTN"]}), False),
        (DataFrame({"invalid": ["ACGTN", "ACGTN"]}), True),
    ],
)
def test_process_data_w_dataframe_raises_appropriate_errors(data, raise_exception):
    model = HyenaDNA()
    if raise_exception:
        with pytest.raises(
            KeyError, match=r"The DataFrame must have a column named 'Sequence'."
        ):
            model.process_data(data)
    else:
        model.process_data(data)



import torch
from torch.autograd import Function


class GradReverse(Function):
    @staticmethod
    def forward(ctx, x: torch.Tensor, lambd: float) -> torch.Tensor:
        ctx.lambd = lambd
        return x.view_as(x)

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor) -> torch.Tensor:
        return grad_output.neg() * ctx.lambd, None


def grad_reverse(x: torch.Tensor, lambd: float = 1.0) -> torch.Tensor:
    return GradReverse.apply(x, lambd)



from typing import Literal, Optional
from helical.models.scgpt.data_collator import DataCollator
from helical.models.scgpt.dataset import Dataset
import torch
from torch import optim
from torch.nn.modules import loss
from torch.utils.data import DataLoader, SequentialSampler
from tqdm import tqdm
from transformers import get_scheduler
from helical.models.base_models import HelicalBaseFineTuningHead
from helical.models.scgpt import scGPT, scGPTConfig
from helical.models.base_models import HelicalBaseFineTuningModel
import logging
import numpy as np

logger = logging.getLogger(__name__)


class scGPTFineTuningModel(HelicalBaseFineTuningModel, scGPT):
    """Fine-tuning model for the scGPT model.

    Example
    ----------
    ```python
    from helical.models.scgpt import scGPTFineTuningModel, scGPTConfig

    # Load the desired dataset
    adata = ad.read_h5ad("dataset.h5ad")

    # Get the desired label class
    cell_types = list(ann_data.obs.cell_type)

    # Get unique labels
    label_set = set(cell_types)

    # Create the fine-tuning model with the relevant configs
    scgpt_config=scGPTConfig(batch_size=10)
    scgpt_fine_tune = scGPTFineTuningModel(scGPT_config=scgpt_config, fine_tuning_head="classification", output_size=len(label_set))

    # Process the data for training
    data = scgpt_fine_tune.process_data(adata)

    # Create a dictionary mapping the classes to unique integers for training
    class_id_dict = dict(zip(label_set, [i for i in range(len(label_set))]))

    for i in range(len(cell_types)):
        cell_types[i] = class_id_dict[cell_types[i]]

    # Fine-tune
    scgpt_fine_tune.train(train_input_data=dataset, train_labels=cell_types)
    ```

    Parameters
    ----------
    scGPT_config : scGPTConfig
        The scGPT configs for fine-tuning model, the same configs that would be used to instantiate the standard scGPT model.
    fine_tuning_head : Literal["classification", "regression"] | HelicalBaseFineTuningHead
        The fine-tuning head that is appended to the model. This can either be a string (options available: "classification", "regression") specifying the task or a custom fine-tuning head inheriting from HelicalBaseFineTuningHead.
    output_size : Optional[int]
        The output size of the fine-tuning model. This is required if the fine_tuning_head is a string specified task. For a classification task this is number of unique classes.

    Methods
    -------
    train(train_input_data: Dataset, train_labels: np.ndarray, validation_input_data: Optional[Dataset], validation_labels: Optional[np.ndarray], optimizer: optim, optimizer_params: dict, loss_function: loss, epochs: int, lr_scheduler_params: Optional[dict])
        Fine-tunes the scGPT model with different head modules.
    get_outputs(dataset: Dataset) -> np.ndarray
        Get the outputs of the fine-tuned model.

    """

    def __init__(
        self,
        scGPT_config: scGPTConfig,
        fine_tuning_head: Literal["classification"] | HelicalBaseFineTuningHead,
        output_size: Optional[int] = None,
    ):
        HelicalBaseFineTuningModel.__init__(self, fine_tuning_head, output_size)
        scGPT.__init__(self, scGPT_config)

        self.fine_tuning_head.set_dim_size(self.config["embsize"])

    def _forward(
        self,
        input_gene_ids: torch.Tensor,
        data_dict: dict,
        src_key_padding_mask: torch.Tensor,
        use_batch_labels: bool,
        device: torch.device,
    ) -> torch.Tensor:
        """
        Forward method of the fine-tuning model.

        Parameters
        ----------
        input_gene_ids : torch.Tensor
            The input tensor to the fine-tuning model.
        data_dict : dict
            The data dictionary containing the expression data and batch labels.
        src_key_padding_mask : torch.Tensor
            The source key padding mask tensor.
        use_batch_labels : bool
            Whether to use batch labels.
        device : torch.device
            The device to run the model on.

        Returns
        -------
        torch.Tensor
            The output tensor of the fine-tuning model.
        """
        embeddings = self.model._encode(
            input_gene_ids,
            data_dict["expr"].to(device),
            src_key_padding_mask=src_key_padding_mask,
            batch_labels=(
                data_dict["batch_labels"].to(device) if use_batch_labels else None
            ),
        )

        if self.config["emb_mode"] == "cls":
            embeddings = embeddings[:, 0, :]
        else:
            embeddings = embeddings[:, 1:, :].mean(dim=1)

        output = self.fine_tuning_head(embeddings)
        return output

    def train(
        self,
        train_input_data: Dataset,
        train_labels: np.ndarray,
        validation_input_data=None,
        validation_labels=None,
        optimizer: optim = optim.AdamW,
        optimizer_params: dict = {"lr": 0.0001},
        loss_function: loss = loss.CrossEntropyLoss(),
        epochs: int = 1,
        # freeze_layers: int = 0,
        lr_scheduler_params: Optional[dict] = None,
    ):
        """Fine-tunes the scGPT model with different head modules.

        Parameters
        ----------
        train_input_data : Dataset
            A helical scGPT processed dataset for fine-tuning
        train_labels : ndarray
            The labels for the training data. These should be stored as unique per class integers.
        validation_input_data : Dataset, default = None
            A helical scGPT processed dataset for per epoch validation. If this is not specified, no validation will be performed.
        validation_labels : ndarray, default = None
            The labels for the validation data. These should be stored as unique per class integers.
        optimizer : torch.optim, default = torch.optim.AdamW
            The optimizer to be used for training.
        optimizer_params : dict
            The optimizer parameters to be used for the optimizer specified. This list should NOT include model parameters.
            e.g. optimizer_params = {'lr': 0.0001}
        loss_function : torch.nn.modules.loss, default = torch.nn.modules.loss.CrossEntropyLoss()
            The loss function to be used.
        epochs : int, optional, default = 10
            The number of epochs to train the model
        lr_scheduler_params : dict, default = None
            The learning rate scheduler parameters for the transformers get_scheduler method. The optimizer will be taken from the optimizer input and should not be included in the learning scheduler parameters. If not specified, no scheduler will be used.
            e.g. lr_scheduler_params = { 'name': 'linear', 'num_warmup_steps': 0, 'num_training_steps': 5 }
        """

        device = next(self.model.parameters()).device

        try:
            use_batch_labels = train_input_data.batch_ids is not None
        except:
            use_batch_labels = False

        collator = DataCollator(
            do_padding=True,
            pad_token_id=self.vocab[self.config["pad_token"]],
            pad_value=self.config["pad_value"],
            do_mlm=False,
            do_binning=True,
            max_length=1200,
            sampling=True,
            keep_first_n_tokens=1,
        )

        data_loader = DataLoader(
            train_input_data,
            batch_size=self.config["batch_size"],
            sampler=SequentialSampler(train_input_data),
            collate_fn=collator,
            drop_last=False,
            pin_memory=True,
        )

        if validation_input_data is not None:
            validation_data_loader = DataLoader(
                validation_input_data,
                batch_size=self.config["batch_size"],
                sampler=SequentialSampler(validation_input_data),
                collate_fn=collator,
                drop_last=False,
                pin_memory=True,
            )

        self.to(device)
        self.model.train()
        self.fine_tuning_head.train()
        optimizer = optimizer(self.parameters(), **optimizer_params)

        lr_scheduler = None
        if lr_scheduler_params is not None:
            lr_scheduler = get_scheduler(optimizer=optimizer, **lr_scheduler_params)

        logger.info("Starting Fine-Tuning")
        for j in range(epochs):
            batch_count = 0
            batch_loss = 0.0
            batches_processed = 0
            training_loop = tqdm(data_loader)
            for data_dict in training_loop:
                input_gene_ids = data_dict["gene"].to(device)
                src_key_padding_mask = input_gene_ids.eq(
                    self.vocab[self.config["pad_token"]]
                )
                output = self._forward(
                    input_gene_ids,
                    data_dict,
                    src_key_padding_mask,
                    use_batch_labels,
                    device,
                )
                labels = torch.tensor(
                    train_labels[batch_count : batch_count + self.config["batch_size"]],
                    device=device,
                )
                batch_count += self.config["batch_size"]
                loss = loss_function(output, labels)
                loss.backward()
                batch_loss += loss.item()
                batches_processed += 1
                optimizer.step()
                optimizer.zero_grad()

                training_loop.set_postfix({"loss": batch_loss / batches_processed})
                training_loop.set_description(f"Fine-Tuning: epoch {j+1}/{epochs}")

            if lr_scheduler is not None:
                lr_scheduler.step()

            if validation_input_data is not None:
                testing_loop = tqdm(
                    validation_data_loader, desc="Fine-Tuning Validation"
                )
                val_loss = 0.0
                count = 0.0
                validation_batch_count = 0
                for validation_data_dict in testing_loop:
                    input_gene_ids = validation_data_dict["gene"].to(device)
                    src_key_padding_mask = input_gene_ids.eq(
                        self.vocab[self.config["pad_token"]]
                    )
                    output = self._forward(
                        input_gene_ids,
                        validation_data_dict,
                        src_key_padding_mask,
                        use_batch_labels,
                        device,
                    )
                    val_labels = torch.tensor(
                        validation_labels[
                            validation_batch_count : validation_batch_count
                            + self.config["batch_size"]
                        ],
                        device=device,
                    )
                    val_loss += loss_function(output, val_labels).item()
                    validation_batch_count += self.config["batch_size"]
                    count += 1.0
                    testing_loop.set_postfix({"val_loss": val_loss / count})
        logger.info(f"Fine-Tuning Complete. Epochs: {epochs}")

    def get_outputs(
        self,
        dataset: Dataset,
    ) -> np.ndarray:
        """Get the outputs of the fine-tuned model.

        Parameters
        ----------
        dataset : Dataset
            The dataset to get the outputs from.

        Returns
        -------
        np.ndarray
            The outputs of the fine-tuned model.
        """
        device = next(self.model.parameters()).device
        self.to(device)
        self.model.eval()
        self.fine_tuning_head.eval()
        
        # fix seeds
        np.random.seed(self.config["binning_seed"])
        torch.manual_seed(self.config["binning_seed"])

        try:
            use_batch_labels = dataset.batch_ids is not None
        except:
            use_batch_labels = False

        collator = DataCollator(
            do_padding=True,
            pad_token_id=self.vocab[self.config["pad_token"]],
            pad_value=self.config["pad_value"],
            do_mlm=False,
            do_binning=True,
            max_length=1200,
            sampling=True,
            keep_first_n_tokens=1,
        )

        data_loader = DataLoader(
            dataset,
            batch_size=self.config["batch_size"],
            sampler=SequentialSampler(dataset),
            collate_fn=collator,
            drop_last=False,
            pin_memory=True,
        )

        testing_loop = tqdm(data_loader, desc="Fine-Tuning Validation")
        outputs = []
        for validation_data_dict in testing_loop:
            input_gene_ids = validation_data_dict["gene"].to(device)
            src_key_padding_mask = input_gene_ids.eq(
                self.vocab[self.config["pad_token"]]
            )
            output = self._forward(
                input_gene_ids,
                validation_data_dict,
                src_key_padding_mask,
                use_batch_labels,
                device,
            )
            outputs.append(output.detach().cpu().numpy())

        return np.vstack(outputs)



import logging
import sys

logger = logging.getLogger("c2s")
# check if logger has been initialized
if not logger.hasHandlers() or len(logger.handlers) == 0:
    logger.propagate = False
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(name)s - %(levelname)s - %(message)s", datefmt="%H:%M:%S"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

from .model import Cell2Sen
from .config import Cell2SenConfig



from helical.models.base_models import HelicalRNAModel
from helical.models.mamba2_mrna.mamba2_mrna_config import Mamba2mRNAConfig
from .mamba2_mrna_tokenizer import CharTokenizer
from datasets import Dataset
from transformers import Mamba2Model, Mamba2Config, BatchEncoding
import torch
from torch.utils.data import DataLoader
import numpy as np
from tqdm import tqdm
from typing import Union
from pandas import DataFrame

import logging

LOGGER = logging.getLogger(__name__)


class Mamba2mRNA(HelicalRNAModel):
    """Mamba2-mRNA Model.

    The Mamba2-mRNA Model is a transformer-based model that can be used to extract mRNA embeddings from mRNA sequences.
    The model is based on the Mamba2 model, which is a transformer-based model trained on mRNA sequences. The model is available through this interface.

    Example
    ----------
    ```python
    from helical.models.mamba2_mrna import Mamba2mRNA, Mamba2mRNAConfig
    import torch

    device = "cuda" if torch.cuda.is_available() else "cpu"

    input_sequences = ["ACUG"*20, "AUGC"*20, "AUGC"*20, "ACUG"*20, "AUUG"*20]

    mamba2_mrna_config = Mamba2mRNAConfig(batch_size=5, device=device)
    mamba2_mrna = Mamba2mRNA(configurer=mamba2_mrna_config)

    processed_input_data = mamba2_mrna.process_data(input_sequences)

    embeddings = mamba2_mrna.get_embeddings(processed_input_data)
    print(embeddings.shape)
    ```


    Parameters
    ----------
    configurer : Mamba2mRNAConfig
        The configuration object for the Mamba2-mRNA model.
    """

    default_configurer = Mamba2mRNAConfig()

    def __init__(self, configurer: Mamba2mRNAConfig = default_configurer):
        super().__init__()
        self.configurer = configurer
        self.config = configurer.config

        self.model = Mamba2Model.from_pretrained(self.config["model_name"])
        self.pretrained_config = Mamba2Config.from_pretrained(
            self.config["model_name"], trust_remote=True
        )
        self.tokenizer = CharTokenizer.from_pretrained(
            self.config["model_name"], trust_remote=True
        )
        self.model.to(self.config["device"])
        self.model.post_init()

        LOGGER.info("Mamba2-mRNA initialized successfully.")
        mode = "training" if self.model.training else "eval"
        LOGGER.info(
            f"'{self.config['model_name']}' model is in '{mode}' mode, on device '{next(self.model.parameters()).device.type}'."
        )

    def process_data(self, sequences: Union[list[str], DataFrame]) -> Dataset:
        """Process the mRNA sequences and return a Dataset object.

        Parameters
        ----------
        sequences : str or DataFrame
            The mRNA sequences. If a DataFrame is provided, it should have a column named "Sequence".

        Returns
        -------
        Dataset
            The dataset object.
        """
        LOGGER.info(f"Processing data for Mamba2-mRNA.")
        sequences = self.get_valid_rna_sequence(sequences)

        tokenized_sequences = self.tokenizer(
            sequences,
            return_tensors="pt",
            padding="max_length",
            truncation=True,
            max_length=self.config["input_size"],
            return_special_tokens_mask=True,
        )

        dataset = Dataset.from_dict(tokenized_sequences)

        LOGGER.info(f"Successfully preprocessed the data for Mamba2-mRNA.")
        return dataset

    def get_embeddings(self, dataset: Dataset) -> np.ndarray:
        """Get the embeddings for the mRNA sequences.

        Parameters
        ----------
        dataset : Dataset
            The dataset object.

        Returns
        -------
        np.ndarray
            The embeddings array.
        """
        LOGGER.info(f"Started getting embeddings:")
        dataloader = DataLoader(
            dataset,
            collate_fn=self._collate_fn,
            batch_size=self.config["batch_size"],
            shuffle=False,
        )
        embeddings = []

        progress_bar = tqdm(dataloader, desc="Getting embeddings")
        with torch.no_grad():
            for batch in progress_bar:
                input_ids = batch["input_ids"].to(self.config["device"])
                special_tokens_mask = batch["special_tokens_mask"].to(
                    self.config["device"]
                )
                attention_mask = batch["attention_mask"].to(self.config["device"])

                output = self.model(
                    input_ids,
                    special_tokens_mask=special_tokens_mask,
                    attention_mask=attention_mask,
                )

                last_hidden_states = output[0]

                if input_ids is not None:
                    batch_size, _ = input_ids.shape[:2]

                if self.pretrained_config.pad_token_id is None and batch_size > 1:
                    message = (
                        "Cannot handle batch sizes > 1 if no padding token is defined."
                    )
                    LOGGER.error(message)
                    raise ValueError(message)

                if self.pretrained_config.pad_token_id is None:
                    sequence_lengths = -1
                else:
                    if input_ids is not None:
                        sequence_lengths = (
                            torch.eq(input_ids, self.pretrained_config.pad_token_id)
                            .int()
                            .argmax(-1)
                            - 1
                        )
                        sequence_lengths = sequence_lengths % input_ids.shape[-1]
                        sequence_lengths = sequence_lengths.to(
                            last_hidden_states.device
                        )
                    else:
                        sequence_lengths = -1

                pooled_last_hidden_states = last_hidden_states[
                    torch.arange(batch_size, device=last_hidden_states.device),
                    sequence_lengths,
                ]

                # Take second last element from the output as last element is a special token
                embeddings.append(pooled_last_hidden_states.cpu().numpy())

                del batch
                del output

        LOGGER.info(f"Finished getting embeddings.")
        return np.concatenate(embeddings)

    def _collate_fn(self, batch):
        input_ids = torch.tensor([item["input_ids"] for item in batch])
        special_tokens_mask = torch.tensor(
            [item["special_tokens_mask"] for item in batch]
        )

        batch_dict = {
            "input_ids": input_ids,
            "special_tokens_mask": special_tokens_mask,
            "attention_mask": 1 - special_tokens_mask,
        }

        if "labels" in batch[0]:
            batch_dict["labels"] = torch.tensor([item["labels"] for item in batch])

        return BatchEncoding(batch_dict)




import hydra
from omegaconf import DictConfig
import anndata as ad
from helical.models.c2s import Cell2Sen, Cell2SenConfig

@hydra.main(version_base=None, config_path="configs", config_name="c2s_config")
def run(cfg: DictConfig):

    adata = ad.read_h5ad("./yolksac_human.h5ad")
    n_cells = 10
    n_genes = 200
    adata = adata[:n_cells, :n_genes].copy()
    perturbation_column = "perturbation"
    adata.obs[perturbation_column] = ["IFNg"] * n_cells

    config = Cell2SenConfig(**cfg)
    c2s = Cell2Sen(configurer=config)

    processed_dataset = c2s.process_data(adata)
    embeddings = c2s.get_embeddings(processed_dataset)
    perturbed_dataset, perturbed_cell_sentences = c2s.get_perturbations(processed_dataset)

if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
"""HyenaDNA training & inference example (Public)

This code is adapted from the original colab tutorial on HyenaDNA. Check that out for an easier entry point into the code.

We provide the code here as an example for those who want something outside collab, with Huggingface integration.

Original file is located at
    https://colab.research.google.com/drive/1wyVEQd4R3HYLTUOXEEQmp_I8aNC_aLhL


"""

# @title Imports
# for HyenaDNA specifically
import torch
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from functools import partial
from einops import rearrange
from typing import Optional
from functools import partial
from torch import Tensor
from helical.models.hyena_dna.stochastic_depth import StochasticDepth
from collections import namedtuple
import numpy as np
import os
import json
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

from transformers.tokenization_utils import AddedToken, PreTrainedTokenizer


"""# HyenaDNA


"""

# @title Hyena layer


def fftconv(u, k, D):
    """
    We apply a convolution through the fourier domain (from the Convolution Theorem)

    """
    seqlen = u.shape[-1]
    fft_size = 2 * seqlen

    k_f = torch.fft.rfft(k, n=fft_size) / fft_size
    u_f = torch.fft.rfft(u.to(dtype=k.dtype), n=fft_size)

    if len(u.shape) > 3:
        k_f = k_f.unsqueeze(1)
    y = torch.fft.irfft(u_f * k_f, n=fft_size, norm="forward")[..., :seqlen]

    out = y + u * D.unsqueeze(-1)
    return out.to(dtype=u.dtype)


@torch.jit.script
def mul_sum(q, y):
    return (q * y).sum(dim=1)


class OptimModule(nn.Module):
    """Interface for Module that allows registering buffers/parameters with configurable optimizer hyperparameters"""

    def register(self, name, tensor, lr=None, wd=0.0):
        """Register a tensor with a configurable learning rate and 0 weight decay"""

        if lr == 0.0:
            self.register_buffer(name, tensor)
        else:
            self.register_parameter(name, nn.Parameter(tensor))

            optim = {}
            if lr is not None:
                optim["lr"] = lr
            if wd is not None:
                optim["weight_decay"] = wd
            setattr(getattr(self, name), "_optim", optim)


class Sin(nn.Module):
    """The Sin activation function for the Hyena Filter function."""

    def __init__(self, dim, w=10, train_freq=True):
        super().__init__()
        self.freq = (
            nn.Parameter(w * torch.ones(1, dim))
            if train_freq
            else w * torch.ones(1, dim)
        )

    def forward(self, x):
        return torch.sin(self.freq * x)


class PositionalEmbedding(OptimModule):
    def __init__(self, emb_dim: int, seq_len: int, lr_pos_emb: float = 1e-5, **kwargs):
        """Complex exponential positional embeddings for Hyena filters."""
        super().__init__()

        self.seq_len = seq_len
        # The time embedding fed to the filteres is normalized so that t_f = 1
        t = torch.linspace(0, 1, self.seq_len)[None, :, None]  # 1, L, 1

        if emb_dim > 1:
            bands = (emb_dim - 1) // 2
        # To compute the right embeddings we use the "proper" linspace
        t_rescaled = torch.linspace(0, seq_len - 1, seq_len)[None, :, None]
        w = 2 * math.pi * t_rescaled / seq_len  # 1, L, 1

        f = torch.linspace(1e-4, bands - 1, bands)[None, None]
        z = torch.exp(-1j * f * w)
        z = torch.cat([t, z.real, z.imag], dim=-1)
        self.register("z", z, lr=lr_pos_emb)
        self.register("t", t, lr=0.0)

    def forward(self, L):
        return self.z[:, :L], self.t[:, :L]


class ExponentialModulation(OptimModule):
    """The window function applied to the output of the (MLP) filter function."""

    def __init__(
        self,
        d_model,
        fast_decay_pct=0.3,
        slow_decay_pct=1.5,
        target=1e-2,
        modulation_lr=0.0,
        modulate: bool = True,
        shift: float = 0.05,
        **kwargs,
    ):
        super().__init__()
        self.modulate = modulate
        self.shift = shift
        max_decay = math.log(target) / fast_decay_pct
        min_decay = math.log(target) / slow_decay_pct
        deltas = torch.linspace(min_decay, max_decay, d_model)[None, None]
        self.register("deltas", deltas, lr=modulation_lr)

    def forward(self, t, x):
        if self.modulate:
            decay = torch.exp(-t * self.deltas.abs())
            x = x * (decay + self.shift)
        return x


class HyenaFilter(OptimModule):
    def __init__(
        self,
        d_model,
        emb_dim=3,  # dim of input to MLP, augments with positional encoding
        order=16,  # width of the implicit MLP
        fused_fft_conv=False,
        seq_len=1024,
        lr=1e-3,
        lr_pos_emb=1e-5,
        dropout=0.0,
        w=1,  # frequency of periodic activations
        wd=0,  # weight decay of kernel parameters
        bias=True,
        num_inner_mlps=2,
        normalized=False,
        **kwargs,
    ):
        """
        Implicit long filter with modulation.

        Args:
            d_model: number of channels in the input
            emb_dim: dimension of the positional encoding (`emb_dim` - 1) // 2 is the number of bands
            order: width of the FFN
            num_inner_mlps: number of inner linear layers inside filter MLP

        Note:
            filter_dropout is not implemented
        """
        super().__init__()

        self.d_model = d_model
        self.use_bias = bias
        self.fused_fft_conv = fused_fft_conv
        self.bias = nn.Parameter(torch.randn(self.d_model))
        self.dropout = nn.Dropout(dropout)

        act = Sin(dim=order, w=w)
        self.emb_dim = emb_dim
        assert (
            emb_dim % 2 != 0 and emb_dim >= 3
        ), "emb_dim must be odd and greater or equal to 3 (time, sine and cosine)"
        self.seq_len = seq_len

        self.pos_emb = PositionalEmbedding(emb_dim, seq_len, lr_pos_emb)

        self.implicit_filter = nn.Sequential(
            nn.Linear(emb_dim, order),
            act,
        )
        for i in range(num_inner_mlps):
            self.implicit_filter.append(nn.Linear(order, order))
            self.implicit_filter.append(act)

        self.implicit_filter.append(nn.Linear(order, d_model, bias=False))

        self.modulation = ExponentialModulation(d_model, **kwargs)

        self.normalized = normalized
        for c in self.implicit_filter.children():
            for name, v in c.state_dict().items():
                optim = {"weight_decay": wd, "lr": lr}
                setattr(getattr(c, name), "_optim", optim)

    def filter(self, L, *args, **kwargs):
        z, t = self.pos_emb(L)
        h = self.implicit_filter(z)
        h = self.modulation(t, h)
        return h

    def forward(self, x, L, k=None, bias=None, *args, **kwargs):
        if k is None:
            k = self.filter(L)

        # Ensure compatibility with filters that return a tuple
        k = k[0] if type(k) is tuple else k

        y = fftconv(x, k, bias)
        return y


class HyenaOperator(nn.Module):
    def __init__(
        self,
        d_model,
        l_max,
        order=2,
        filter_order=64,
        dropout=0.0,
        filter_dropout=0.0,
        **filter_args,
    ):
        r"""
        Hyena operator described in the paper https://arxiv.org/pdf/2302.10866.pdf

        Args:
            d_model (int): Dimension of the input and output embeddings (width of the layer)
            l_max: (int): Maximum input sequence length. Defaults to None
            order: (int): Depth of the Hyena recurrence. Defaults to 2
            dropout: (float): Dropout probability. Defaults to 0.0
            filter_dropout: (float): Dropout probability for the filter. Defaults to 0.0
        """
        super().__init__()

        self.d_model = d_model
        self.l_max = l_max
        self.order = order
        inner_width = d_model * (order + 1)
        self.dropout = nn.Dropout(dropout)
        self.in_proj = nn.Linear(d_model, inner_width)
        self.out_proj = nn.Linear(d_model, d_model)

        self.short_filter = nn.Conv1d(
            inner_width, inner_width, 3, padding=2, groups=inner_width
        )
        self.filter_fn = HyenaFilter(
            d_model * (order - 1),
            order=filter_order,
            seq_len=l_max,
            channels=1,
            dropout=filter_dropout,
            **filter_args,
        )

    def forward(self, u, *args, **kwargs):
        l = u.size(-2)
        l_filter = min(l, self.l_max)
        u = self.in_proj(u)
        u = rearrange(u, "b l d -> b d l")

        uc = self.short_filter(u)[..., :l_filter]
        *x, v = uc.split(self.d_model, dim=1)

        k = self.filter_fn.filter(l_filter)[0]
        k = rearrange(k, "l (o d) -> o d l", o=self.order - 1)
        bias = rearrange(self.filter_fn.bias, "(o d) -> o d", o=self.order - 1)

        for o, x_i in enumerate(reversed(x[1:])):
            v = self.dropout(v * x_i)
            v = self.filter_fn(v, l_filter, k=k[o], bias=bias[o])

        y = rearrange(v * x[0], "b d l -> b l d")

        y = self.out_proj(y)
        return y


# @title Self-Attention (alternative)

"""
If you'd like to try the HyenaDNA model using attention instead, you can. ie,
use a regular decoder only Transformer.
"""


class SelfAttention(nn.Module):
    """Implement the scaled dot product attention with softmax.
    Arguments
    ---------
        softmax_scale: The temperature to use for the softmax attention.
                      (default: 1/sqrt(d_keys) where d_keys is computed at
                      runtime)
        attention_dropout: The dropout rate to apply to the attention
                           (default: 0.0)
    """

    def __init__(self, causal=False, softmax_scale=None, attention_dropout=0.0):
        super().__init__()
        self.causal = causal
        self.softmax_scale = softmax_scale
        self.dropout_p = attention_dropout

    def forward(self, qkv, causal=None, key_padding_mask=None):
        """Implements the multihead softmax attention.
        Arguments
        ---------
            qkv: The tensor containing the query, key, and value. (B, S, 3, H, D)
            causal: if passed, will override self.causal
            key_padding_mask: boolean mask to apply to the attention weights. True means to keep,
                False means to mask out. (B, S)
        """
        batch_size, seqlen = qkv.shape[0], qkv.shape[1]
        causal = self.causal if causal is None else causal
        q, k, v = qkv.unbind(dim=2)
        softmax_scale = self.softmax_scale or 1.0 / math.sqrt(q.shape[-1])
        scores = torch.einsum("bthd,bshd->bhts", q, k * softmax_scale)
        if key_padding_mask is not None:
            padding_mask = torch.full(
                (batch_size, seqlen), -10000.0, dtype=scores.dtype, device=scores.device
            )
            padding_mask.masked_fill_(key_padding_mask, 0.0)
            # TD [2022-09-30]: Adding is faster than masked_fill_ (idk why, just better kernel I guess)
            scores = scores + rearrange(padding_mask, "b s -> b 1 1 s")
        if causal:
            # "triu_tril_cuda_template" not implemented for 'BFloat16'
            # So we have to construct the mask in float
            causal_mask = torch.triu(
                torch.full((seqlen, seqlen), -10000.0, device=scores.device), 1
            )
            # TD [2022-09-30]: Adding is faster than masked_fill_ (idk why, just better kernel I guess)
            scores = scores + causal_mask.to(dtype=scores.dtype)
        attention = torch.softmax(scores, dim=-1, dtype=v.dtype)
        attention_drop = F.dropout(attention, self.dropout_p if self.training else 0.0)
        output = torch.einsum("bhts,bshd->bthd", attention_drop, v)
        return output


class MHA(nn.Module):
    """Multi-head self-attention and cross-attention"""

    def __init__(
        self,
        embed_dim,
        num_heads,
        bias=True,
        dropout=0.0,
        softmax_scale=None,
        causal=False,
        layer_idx=None,
        dwconv=False,
        return_residual=False,
        device=None,
        dtype=None,
    ) -> None:
        """
        return_residual: whether to return the input x along with the output. This is for
            performance reason: for post-norm architecture, returning the input allows us
            to fuse the backward of nn.Linear with the residual connection.
        """
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.embed_dim = embed_dim
        self.causal = causal
        self.layer_idx = layer_idx
        self.dwconv = dwconv
        self.return_residual = return_residual

        self.num_heads = num_heads
        assert (
            self.embed_dim % num_heads == 0
        ), "self.kdim must be divisible by num_heads"
        self.head_dim = self.embed_dim // num_heads

        linear_cls = nn.Linear
        linear_resid_cls = LinearResidual
        inner_attn_cls = SelfAttention

        if not self.return_residual:
            self.Wqkv = linear_cls(
                embed_dim, 3 * embed_dim, bias=bias, **factory_kwargs
            )
        else:
            self.Wqkv = linear_resid_cls(
                embed_dim, 3 * embed_dim, bias=bias, **factory_kwargs
            )
        if self.dwconv:
            self.dwconv_qkv = nn.Conv1d(
                3 * embed_dim,
                3 * embed_dim,
                kernel_size=3,
                padding=2,
                groups=3 * embed_dim,
            )

        self.inner_attn = inner_attn_cls(
            causal=causal, softmax_scale=softmax_scale, attention_dropout=dropout
        )

        # output projection always have the bias (for now)
        self.out_proj = linear_cls(embed_dim, embed_dim, **factory_kwargs)

    def forward(self, x, key_padding_mask=None, **kwargs):
        """
        Arguments:
            x: (batch, seqlen, hidden_dim) (where hidden_dim = num heads * head dim) if
                cu_seqlens is None and max_seqlen is None, else (total, hidden_dim) where total
                is the is the sum of the sequence lengths in the batch.
            cu_seqlens: (batch_size + 1,), dtype torch.int32. The cumulative sequence lengths
                of the sequences in the batch, used to index into x. Only applicable when using
                FlashAttention.
            max_seqlen: int. Maximum sequence length in the batch.
            key_padding_mask: boolean mask, True means to keep, False means to mask out.
                (batch, seqlen). Only applicable when not using FlashAttention.
            mixer_subset: for cross-attention only. If not None, will take a subset of x
                before applying the query projection. Useful for e.g., ViT where we only care
                about the CLS token in the last layer.
            inference_params: for generation. Adapted from Megatron-LM (and Apex)
            https://github.com/NVIDIA/apex/blob/3ff1a10f72ec07067c4e44759442329804ac5162/apex/transformer/testing/standalone_transformer_lm.py#L470
        """

        kwargs = {"key_padding_mask": key_padding_mask, **kwargs}

        if not self.return_residual:
            qkv = self.Wqkv(x)
        else:
            qkv, x = self.Wqkv(x)
        if self.dwconv:
            qkv = rearrange(
                self.dwconv_qkv(rearrange(qkv, "b s d -> b d s"))[..., :-2],
                "b d s -> b s d",
            ).contiguous()
        qkv = rearrange(
            qkv, "... (three h d) -> ... three h d", three=3, d=self.head_dim
        )

        context = self.inner_attn(qkv, **kwargs)

        out = self.out_proj(rearrange(context, "... h d -> ... (h d)"))
        return out if not self.return_residual else (out, x)


# @title MLP layer

"""
The MLP layer after the mixer layer (HyenaOperator).
"""


class Mlp(nn.Module):

    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        activation=F.gelu,
        return_residual=False,
        device=None,
        dtype=None,
    ):
        """
        From https://github.com/HazyResearch/flash-attention/blob/main/flash_attn/modules/mlp.py
        """
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.return_residual = return_residual
        self.fc1 = nn.Linear(in_features, hidden_features, **factory_kwargs)
        self.activation = activation
        self.fc2 = nn.Linear(hidden_features, out_features, **factory_kwargs)

    def forward(self, x):
        y = self.fc1(x)
        y = self.activation(y)
        y = self.fc2(y)
        return y if not self.return_residual else (y, x)


# @title Block layer (Hyena + MLP layers)

"""
A block consists of a Mixer layer (Hyena or attention), and a MLP layer.

"""


class LinearResidual(nn.Linear):
    """Wrap nn.Linear to return the residual as well. For compatibility with FusedDense."""

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return super().forward(input), input


class Block(nn.Module):

    def __init__(
        self,
        dim,
        mixer_cls=None,
        mlp_cls=None,
        norm_cls=nn.LayerNorm,
        dropout_cls=nn.Dropout,
        prenorm=True,
        resid_dropout1=0.0,
        resid_dropout2=0.0,
        drop_path1=0.0,
        drop_path2=0.0,
        return_residual=False,
        residual_in_fp32=False,
    ):
        """
        From https://github.com/HazyResearch/flash-attention/blob/main/flash_attn/modules/block.py
        For prenorm=True, this Block has a slightly different structure compared to a regular
        prenorm Transformer block.
        The standard block is: LN -> MHA -> Dropout -> Add -> LN -> MLP -> Dropout -> Add.
        [Ref: https://arxiv.org/abs/2002.04745]
        Here we have: Dropout -> Add -> LN -> MHA -> Dropout -> Add -> LN -> MLP, returning both
        the hidden_states (output of the MLP) and the residual.
        This is for performance reasons, as we can fuse the dropout, add and LayerNorm.
        The residual needs to be provided (except for the very first block).
        For prenorm=False, this Block has the same structure as a regular postnorm Transformer
        block: MHA -> Dropout -> Add -> LN -> MLP -> Dropout -> Add -> LN.
        return_residual: whether each of the sub-layers (mixer and mlp) will return the residual.
        This is for performance reason: for post-norm architecture, returning the input allows us
        to fuse the backward of nn.Linear with the residual connection.
        """
        super().__init__()
        self.prenorm = prenorm
        self.return_residual = return_residual
        self.residual_in_fp32 = residual_in_fp32
        if self.residual_in_fp32:
            assert self.prenorm, "residual_in_fp32 is only compatible with prenorm=True"
        if mixer_cls is None:
            mixer_cls = partial(MHA, num_heads=dim // 64)
        if mlp_cls is None:
            mlp_cls = partial(Mlp, hidden_features=4 * dim)
        self.mixer = mixer_cls()
        self.dropout1 = dropout_cls(resid_dropout1)
        self.drop_path1 = StochasticDepth(drop_path1, mode="row")
        self.norm1 = norm_cls(dim)
        self.mlp = mlp_cls(dim)
        if not isinstance(self.mlp, nn.Identity):
            self.dropout2 = dropout_cls(resid_dropout2)
            self.drop_path2 = StochasticDepth(drop_path2, mode="row")
            self.norm2 = norm_cls(dim)

    def forward(
        self, hidden_states, residual=None, mixer_subset=None, mixer_kwargs=None
    ):
        r"""Pass the input through the encoder layer.
        Args:
            hidden_states: the sequence to the encoder layer (required).
            residual: if postnorm, residual=None, If prenorm, hidden_states = Attn/MLP(LN(residual))
            mixer_subset: for cross-attention only. If not None, will take a subset of x
                before applying the query projection. Useful for e.g., ViT where we only care
                about the CLS token in the last layer.
        """
        if self.prenorm:
            dropped = self.drop_path1(self.dropout1(hidden_states))
            residual = (dropped + residual) if residual is not None else dropped
            hidden_states = self.norm1(residual.to(dtype=self.norm1.weight.dtype))
            if self.residual_in_fp32:
                residual = residual.to(torch.float32)
            if mixer_kwargs is None:
                mixer_kwargs = {}
            if mixer_subset is not None:
                mixer_kwargs["mixer_subset"] = mixer_subset
            hidden_states = self.mixer(hidden_states, **mixer_kwargs)
            if mixer_subset is not None:
                residual = residual[:, mixer_subset]
            if not isinstance(self.mlp, nn.Identity):
                dropped = self.drop_path2(self.dropout2(hidden_states))
                residual = (dropped + residual) if residual is not None else dropped
                hidden_states = self.norm2(residual.to(dtype=self.norm2.weight.dtype))
                if self.residual_in_fp32:
                    residual = residual.to(torch.float32)

                hidden_states = self.mlp(hidden_states)
            return hidden_states, residual
        else:
            assert residual is None
            mixer_out = self.mixer(
                hidden_states, **(mixer_kwargs if mixer_kwargs is not None else {})
            )
            if self.return_residual:  # mixer out is actually a pair here
                mixer_out, hidden_states = mixer_out

            hidden_states = self.norm1(
                (self.drop_path1(self.dropout1(mixer_out)) + hidden_states).to(
                    dtype=self.norm1.weight.dtype
                )
            )

            if not isinstance(self.mlp, nn.Identity):
                mlp_out = self.mlp(hidden_states)
                if self.return_residual:  # mlp out is actually a pair here
                    mlp_out, hidden_states = mlp_out

                hidden_states = self.norm2(
                    (self.drop_path2(self.dropout2(mlp_out)) + hidden_states).to(
                        dtype=self.norm2.weight.dtype
                    )
                )

            return hidden_states


def create_mixer_cls(
    layer=None,
    attn_layer_idx=None,
    attn_cfg=None,
    layer_idx=None,
    device=None,
    dtype=None,
):
    factory_kwargs = {"device": device, "dtype": dtype}
    if attn_layer_idx is not None and layer_idx in attn_layer_idx:
        causal = True if attn_cfg is None else attn_cfg.pop("causal", True)

        mha_cls = MHA

        mixer_cls = partial(
            mha_cls,
            causal=causal,
            layer_idx=layer_idx,
            **(attn_cfg if attn_cfg is not None else {}),
            **factory_kwargs,
        )
    else:
        # mixer_cls = instantiate(registry.layer, layer, partial=True, layer_idx=layer_idx, **factory_kwargs)

        mixer_cls = partial(HyenaOperator, **layer)

    return mixer_cls


def create_mlp_cls(d_model, d_inner=None, device=None, dtype=None):
    factory_kwargs = {"device": device, "dtype": dtype}
    inner_dim = d_inner if d_inner is not None else 4 * d_model

    mlp_cls = partial(
        Mlp,
        hidden_features=inner_dim,
        activation=partial(F.gelu, approximate="tanh"),
        **factory_kwargs,
    )

    return mlp_cls


def create_block(
    d_model,
    d_inner=None,
    layer=None,
    attn_layer_idx=None,
    attn_cfg=None,
    layer_norm_epsilon=1e-5,
    resid_dropout1=0.0,
    resid_dropout2=0.0,
    residual_in_fp32=False,
    layer_idx=None,
    device=None,
    dtype=None,
):
    factory_kwargs = {"device": device, "dtype": dtype}
    mixer_cls = create_mixer_cls(
        layer=layer,
        attn_layer_idx=attn_layer_idx,
        attn_cfg=attn_cfg,
        layer_idx=layer_idx,
        **factory_kwargs,
    )
    mlp_cls = create_mlp_cls(d_model, d_inner=d_inner, **factory_kwargs)
    norm_cls = partial(nn.LayerNorm, eps=layer_norm_epsilon, **factory_kwargs)
    block = Block(
        d_model,
        mixer_cls,
        mlp_cls,
        norm_cls=norm_cls,
        prenorm=True,
        resid_dropout1=resid_dropout1,
        resid_dropout2=resid_dropout2,
        residual_in_fp32=residual_in_fp32,
    )
    block.layer_idx = layer_idx
    return block


# https://github.com/huggingface/transformers/blob/c28d04e9e252a1a099944e325685f14d242ecdcd/src/transformers/models/gpt2/modeling_gpt2.py#L454
def _init_weights(
    module,
    n_layer,
    initializer_range=0.02,
    rescale_prenorm_residual=True,
    glu_act=False,
):
    if isinstance(module, nn.Linear):
        nn.init.normal_(module.weight, std=initializer_range)
        if module.bias is not None:
            nn.init.zeros_(module.bias)
    elif isinstance(module, nn.Embedding):
        nn.init.normal_(module.weight, std=initializer_range)

    if rescale_prenorm_residual:
        # Reinitialize selected weights subject to the OpenAI GPT-2 Paper Scheme:
        #   > A modified initialization which accounts for the accumulation on the residual path with model depth. Scale
        #   > the weights of residual layers at initialization by a factor of 1/√N where N is the # of residual layers.
        #   >   -- GPT-2 :: https://openai.com/blog/better-language-models/
        #
        # Reference (Megatron-LM): https://github.com/NVIDIA/Megatron-LM/blob/main/megatron/model/gpt_model.py
        for name, p in module.named_parameters():
            if name in ["out_proj.weight", "fc2.weight"]:
                # Special Scaled Initialization --> There are 2 Layer Norms per Transformer Block
                nn.init.normal_(
                    p, mean=0.0, std=initializer_range / math.sqrt(2 * n_layer)
                )
            # If using GLU activation for now, we scale the std by 2
            elif name in ["output_linear.0.weight"]:
                # Special Scaled Initialization --> There are 2 Layer Norms per Transformer Block
                if not glu_act:
                    nn.init.normal_(
                        p, mean=0.0, std=initializer_range / math.sqrt(2 * n_layer)
                    )
                else:
                    out_features = p.shape[0]
                    # Multiplying the first half of the matrix by 2 since sigmoid scales it down by 0.5
                    # on average.
                    nn.init.normal_(
                        p[: out_features // 2],
                        mean=0.0,
                        std=initializer_range / math.sqrt(2 * n_layer) * 2,
                    )


# @title Backbone model (stack of blocks)

"""
A backbone model consists of a stack of blocks. If you use attention, then
positional embeddings are included. When using Hyena, then the pos emb
revert to doing nothing.
"""


class GPT2Embeddings(nn.Module):

    def __init__(
        self,
        embed_dim,
        vocab_size,
        max_position_embeddings,
        padding_idx=None,
        word_embed_proj_dim=None,
        device=None,
        dtype=None,
    ):
        """
        If max_position_embeddings <= 0, there's no position embeddings
        If word_embe_proj_dim is not None (e.g., OPT-350m), we embed to that dimension
            the project up to embed_dim
        """
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        if word_embed_proj_dim is None:
            self.word_embeddings = nn.Embedding(
                vocab_size, embed_dim, padding_idx=padding_idx, **factory_kwargs
            )
            self.project_in = None
        else:
            self.word_embeddings = nn.Embedding(
                vocab_size,
                word_embed_proj_dim,
                padding_idx=padding_idx,
                **factory_kwargs,
            )
            self.project_in = nn.Linear(
                word_embed_proj_dim, embed_dim, bias=False, **factory_kwargs
            )
        self.max_position_embeddings = max_position_embeddings
        if self.max_position_embeddings > 0:
            self.position_embeddings = nn.Embedding(
                max_position_embeddings, embed_dim, **factory_kwargs
            )

    def forward(self, input_ids, position_ids=None):
        """
        input_ids: (batch, seqlen)
        position_ids: (batch, seqlen)
        """
        batch_size, seqlen = input_ids.shape
        embeddings = self.word_embeddings(input_ids)
        if self.project_in is not None:
            embeddings = self.project_in(embeddings)
        if self.max_position_embeddings > 0:
            if position_ids is None:
                position_ids = torch.arange(
                    seqlen, dtype=torch.long, device=input_ids.device
                )
            position_embeddings = self.position_embeddings(position_ids)
            embeddings = embeddings + position_embeddings
        return embeddings


class LMBackbone(nn.Module):

    def __init__(
        self,
        d_model: int,
        n_layer: int,
        d_inner: int,
        vocab_size: int,
        process_group=None,
        layer=None,
        attn_layer_idx=None,
        attn_cfg=None,
        max_position_embeddings=0,
        resid_dropout: float = 0.0,
        embed_dropout: float = 0.1,
        layer_norm_epsilon: float = 1e-5,
        initializer_cfg=None,
        residual_in_fp32=False,
        device=None,
        dtype=None,
        **kwargs,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.process_group = process_group
        self.residual_in_fp32 = residual_in_fp32
        # note max_position_embeddings is 0 for Hyena, and therefore isn't used
        self.embeddings = GPT2Embeddings(
            d_model, vocab_size, max_position_embeddings, **factory_kwargs
        )

        self.layers = nn.ModuleList(
            [
                create_block(
                    d_model,
                    d_inner=d_inner,
                    layer=layer,
                    attn_layer_idx=attn_layer_idx,
                    attn_cfg=attn_cfg,
                    layer_norm_epsilon=layer_norm_epsilon,
                    resid_dropout1=embed_dropout if i == 0 else resid_dropout,
                    resid_dropout2=resid_dropout,
                    residual_in_fp32=residual_in_fp32,
                    layer_idx=i,
                    **factory_kwargs,
                )
                for i in range(n_layer)
            ]
        )

        self.drop_f = nn.Dropout(resid_dropout)
        self.ln_f = nn.LayerNorm(d_model, eps=layer_norm_epsilon, **factory_kwargs)

        self.apply(
            partial(
                _init_weights,
                n_layer=n_layer,
                **(initializer_cfg if initializer_cfg is not None else {}),
            )
        )

    def forward(self, input_ids, position_ids=None):
        hidden_states = self.embeddings(
            input_ids,
            position_ids=position_ids,
        )
        residual = None

        for layer in self.layers:
            hidden_states, residual = layer(hidden_states, residual)

        dropped = self.drop_f(hidden_states)
        residual = (dropped + residual) if residual is not None else dropped
        hidden_states = self.ln_f(residual.to(dtype=self.ln_f.weight.dtype))

        return hidden_states


# @title Decoder head layer

"""
A simple decoder head (using MLP) to predict a sequence level classification.
You have the option to average across all the tokens in a sequence or using the
"last" token to classify.  At least, those 2 worked best for us, but we provide
other "modes" as well.

We only need this for classification.  Otherwise we'll use the hidden
states of the backbone as embeddings.

"""


class SequenceDecoder(nn.Module):
    def __init__(
        self, d_model, d_output=None, l_output=None, use_lengths=False, mode="last"
    ):
        super().__init__()

        self.output_transform = (
            nn.Identity() if d_output is None else nn.Linear(d_model, d_output)
        )

        if l_output is None:
            self.l_output = None
            self.squeeze = False
        elif l_output == 0:
            # Equivalent to getting an output of length 1 and then squeezing
            self.l_output = 1
            self.squeeze = True
        else:
            assert l_output > 0
            self.l_output = l_output
            self.squeeze = False

        self.use_lengths = use_lengths
        self.mode = mode

        if mode == "ragged":
            assert not use_lengths

    def forward(self, x, state=None, lengths=None, l_output=None):
        """
        x: (n_batch, l_seq, d_model)
        Returns: (n_batch, l_output, d_output)
        """

        if self.l_output is None:
            if l_output is not None:
                assert isinstance(l_output, int)  # Override by pass in
            else:
                # Grab entire output
                l_output = x.size(-2)
            squeeze = False
        else:
            l_output = self.l_output
            squeeze = self.squeeze

        if self.mode == "last":
            restrict = lambda x: x[..., -l_output:, :]
        elif self.mode == "first":
            restrict = lambda x: x[..., :l_output, :]
        elif self.mode == "pool":
            restrict = lambda x: (
                torch.cumsum(x, dim=-2)
                / torch.arange(
                    1, 1 + x.size(-2), device=x.device, dtype=x.dtype
                ).unsqueeze(-1)
            )[..., -l_output:, :]

            def restrict(x):
                L = x.size(-2)
                s = x.sum(dim=-2, keepdim=True)
                if l_output > 1:
                    c = torch.cumsum(x[..., -(l_output - 1) :, :].flip(-2), dim=-2)
                    c = F.pad(c, (0, 0, 1, 0))
                    s = s - c  # (B, l_output, D)
                    s = s.flip(-2)
                denom = torch.arange(
                    L - l_output + 1, L + 1, dtype=x.dtype, device=x.device
                )
                s = s / denom
                return s

        elif self.mode == "sum":
            restrict = lambda x: torch.cumsum(x, dim=-2)[..., -l_output:, :]
            # TODO use same restrict function as pool case
        elif self.mode == "ragged":
            assert lengths is not None, "lengths must be provided for ragged mode"
            # remove any additional padding (beyond max length of any sequence in the batch)
            restrict = lambda x: x[..., : max(lengths), :]
        else:
            raise NotImplementedError(
                "Mode must be ['last' | 'first' | 'pool' | 'sum']"
            )

        # Restrict to actual length of sequence
        if self.use_lengths:
            assert lengths is not None
            x = torch.stack(
                [
                    restrict(out[..., :length, :])
                    for out, length in zip(torch.unbind(x, dim=0), lengths)
                ],
                dim=0,
            )
        else:
            x = restrict(x)

        if squeeze:
            assert x.size(-2) == 1
            x = x.squeeze(-2)

        x = self.output_transform(x)

        return x

    def step(self, x, state=None):
        # Ignore all length logic
        return self.output_transform(x)


# @title Model (backbone + head)

"""
Putting it all together, the model consists of a backbone model
and a decoder head (you can turn off head for embeddings only too).

Here we use a simple head to do multi-classification, but
can also swap the head to do next token prediction too.  We defer to the main
HyenaDNA for that code, since pretraining with next token prediction isn't quite
feasible on colab.

"""


class HyenaDNAModel(nn.Module):

    def __init__(
        self,
        d_model: int,
        n_layer: int,
        d_inner: int,
        vocab_size: int,
        layer=None,
        attn_layer_idx=None,
        attn_cfg=None,
        max_position_embeddings=0,
        resid_dropout: float = 0.0,
        embed_dropout: float = 0.1,
        layer_norm_epsilon: float = 1e-5,
        initializer_cfg=None,
        residual_in_fp32=False,
        pad_vocab_size_multiple: int = 1,
        use_head=False,
        n_classes: int = 2,
        device=None,
        dtype=None,
        **kwargs,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        if vocab_size % pad_vocab_size_multiple != 0:
            vocab_size += pad_vocab_size_multiple - (
                vocab_size % pad_vocab_size_multiple
            )

        self.use_head = use_head

        layer["d_model"] = d_model

        self.backbone = LMBackbone(
            d_model=d_model,
            n_layer=n_layer,
            d_inner=d_inner,
            vocab_size=vocab_size,
            layer=layer,
            attn_layer_idx=attn_layer_idx,
            attn_cfg=attn_cfg,
            max_position_embeddings=max_position_embeddings,
            resid_dropout=resid_dropout,
            embed_dropout=embed_dropout,
            layer_norm_epsilon=layer_norm_epsilon,
            initializer_cfg=initializer_cfg,
            residual_in_fp32=residual_in_fp32,
            **factory_kwargs,
            **kwargs,
        )

        # we only need a head if doing classification, otherwise we'll use the
        # hidden states as embeddings
        if self.use_head:
            self.head = SequenceDecoder(
                d_model=d_model, d_output=n_classes, l_output=0, mode="pool"
            )

        # Initialize weights and apply final processing
        self.apply(
            partial(
                _init_weights,
                n_layer=n_layer,
                **(initializer_cfg if initializer_cfg is not None else {}),
            )
        )

        # if self.use_head:
        #     self.tie_weights()

    # def tie_weights(self):
    #     self.head.weight = self.backbone.embeddings.word_embeddings.weight

    def forward(
        self, input_ids, position_ids=None, state=None
    ):  # state for the repo interface
        hidden_states = self.backbone(input_ids, position_ids=position_ids)

        if self.use_head:
            return self.head(hidden_states)
        else:
            return hidden_states


"""# Data pipeline


"""

# @title Tokenizer

"""
Just a simple character level tokenizer.

From: https://github.com/dariush-bahrami/character-tokenizer/blob/master/charactertokenizer/core.py

CharacterTokenzier for Hugging Face Transformers.
This is heavily inspired from CanineTokenizer in transformers package.
"""


class CharacterTokenizer(PreTrainedTokenizer):
    def __init__(
        self,
        characters: Sequence[str],
        model_max_length: int,
        padding_side: str = "left",
        **kwargs,
    ):
        """Character tokenizer for Hugging Face transformers.
        Args:
            characters (Sequence[str]): List of desired characters. Any character which
                is not included in this list will be replaced by a special token called
                [UNK] with id=6. Following are list of all of the special tokens with
                their corresponding ids:
                    "[CLS]": 0
                    "[SEP]": 1
                    "[BOS]": 2
                    "[MASK]": 3
                    "[PAD]": 4
                    "[RESERVED]": 5
                    "[UNK]": 6
                an id (starting at 7) will be assigned to each character.
            model_max_length (int): Model maximum sequence length.
        """
        self.characters = characters
        self.model_max_length = model_max_length
        bos_token = AddedToken("[BOS]", lstrip=False, rstrip=False)
        eos_token = AddedToken("[SEP]", lstrip=False, rstrip=False)
        sep_token = AddedToken("[SEP]", lstrip=False, rstrip=False)
        cls_token = AddedToken("[CLS]", lstrip=False, rstrip=False)
        pad_token = AddedToken("[PAD]", lstrip=False, rstrip=False)
        unk_token = AddedToken("[UNK]", lstrip=False, rstrip=False)

        mask_token = AddedToken("[MASK]", lstrip=True, rstrip=False)

        self._vocab_str_to_int = {
            "[CLS]": 0,
            "[SEP]": 1,
            "[BOS]": 2,
            "[MASK]": 3,
            "[PAD]": 4,
            "[RESERVED]": 5,
            "[UNK]": 6,
            **{ch: i + 7 for i, ch in enumerate(characters)},
        }
        self._vocab_int_to_str = {v: k for k, v in self._vocab_str_to_int.items()}

        super().__init__(
            bos_token=bos_token,
            eos_token=sep_token,
            sep_token=sep_token,
            cls_token=cls_token,
            pad_token=pad_token,
            mask_token=mask_token,
            unk_token=unk_token,
            add_prefix_space=False,
            model_max_length=model_max_length,
            padding_side=padding_side,
            **kwargs,
        )

    @property
    def vocab_size(self) -> int:
        return len(self._vocab_str_to_int)

    def _tokenize(self, text: str) -> List[str]:
        return list(text)

    def _convert_token_to_id(self, token: str) -> int:
        return self._vocab_str_to_int.get(token, self._vocab_str_to_int["[UNK]"])

    def _convert_id_to_token(self, index: int) -> str:
        return self._vocab_int_to_str[index]

    def convert_tokens_to_string(self, tokens):
        return "".join(tokens)

    def build_inputs_with_special_tokens(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        sep = [self.sep_token_id]
        cls = [self.cls_token_id]
        result = cls + token_ids_0 + sep
        if token_ids_1 is not None:
            result += token_ids_1 + sep
        return result

    def get_special_tokens_mask(
        self,
        token_ids_0: List[int],
        token_ids_1: Optional[List[int]] = None,
        already_has_special_tokens: bool = False,
    ) -> List[int]:
        if already_has_special_tokens:
            return super().get_special_tokens_mask(
                token_ids_0=token_ids_0,
                token_ids_1=token_ids_1,
                already_has_special_tokens=True,
            )

        result = [1] + ([0] * len(token_ids_0)) + [1]
        if token_ids_1 is not None:
            result += ([0] * len(token_ids_1)) + [1]
        return result

    def create_token_type_ids_from_sequences(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        sep = [self.sep_token_id]
        cls = [self.cls_token_id]

        result = len(cls + token_ids_0 + sep) * [0]
        if token_ids_1 is not None:
            result += len(token_ids_1 + sep) * [1]
        return result

    def get_config(self) -> Dict:
        return {
            "char_ords": [ord(ch) for ch in self.characters],
            "model_max_length": self.model_max_length,
        }

    @classmethod
    def from_config(cls, config: Dict) -> "CharacterTokenizer":
        cfg = {}
        cfg["characters"] = [chr(i) for i in config["char_ords"]]
        cfg["model_max_length"] = config["model_max_length"]
        return cls(**cfg)

    def save_pretrained(self, save_directory: Union[str, os.PathLike], **kwargs):
        cfg_file = Path(save_directory) / "tokenizer_config.json"
        cfg = self.get_config()
        with open(cfg_file, "w") as f:
            json.dump(cfg, f, indent=4)

    def get_vocab(self) -> Dict[str, int]:
        return self._vocab_str_to_int

    @classmethod
    def from_pretrained(cls, save_directory: Union[str, os.PathLike], **kwargs):
        cfg_file = Path(save_directory) / "tokenizer_config.json"
        with open(cfg_file) as f:
            cfg = json.load(f)
        return cls.from_config(cfg)



from helical.models.mamba2_mrna import Mamba2mRNA, Mamba2mRNAConfig
import hydra
from omegaconf import DictConfig


@hydra.main(
    version_base=None,
    config_path="../run_models/configs",
    config_name="mamba2_mrna_config",
)
def run(cfg: DictConfig):
    input_sequences = ["ACUG" * 20, "AUGC" * 20, "AUGC" * 20, "ACUG" * 20, "AUUG" * 20]

    mamba2_mrna_config = Mamba2mRNAConfig(**cfg)
    mamba2_mrna = Mamba2mRNA(mamba2_mrna_config)

    processed_input_data = mamba2_mrna.process_data(input_sequences)

    embeddings = mamba2_mrna.get_embeddings(processed_input_data)
    print(embeddings.shape)


if __name__ == "__main__":
    run()



from typing import Optional
from pathlib import Path
from helical.constants.paths import CACHE_DIR_HELICAL
from typing import Literal


class GenePTConfig:
    """Configuration class to use the GenePT Model.

    Parameters
    ----------
    model_name : Literal["gpt3.5"], optional, default="gpt3.5"
        The name of the model for the embeddings.
    batch_size : int, optional, default = 24
        The batch size
    emb_layer : int, optional, default = -1
        The embedding layer
    emb_mode : Literal["cls", "cell", "gene"], optional, default="cell"
        The embedding mode
    device : Literal["cpu", "cuda"], optional, default="cpu"
        The device to use. Either use "cuda" or "cpu".
    accelerator : bool, optional, default=False
        The accelerator configuration. By default same device as model.
    nproc: int, optional, default=1
        Number of processes to use for data processing.
    custom_attr_name_dict : dict, optional, default=None
        A dictionary that contains the names of the custom attributes to be added to the dataset.
        The keys of the dictionary are the names of the custom attributes, and the values are the names of the columns in adata.obs.
        For example, if you want to add a custom attribute called "cell_type" to the dataset, you would pass custom_attr_name_dict = {"cell_type": "cell_type"}.
        If you do not want to add any custom attributes, you can leave this parameter as None.
    Returns
    -------
    GenePTConfig
        The GenePT configuration object

    """

    def __init__(
        self,
        model_name: Literal["gpt3.5"] = "gpt3.5",
        batch_size: int = 24,
        emb_layer: int = -1,
        emb_mode: Literal["cls", "cell", "gene"] = "cell",
        device: Literal["cpu", "cuda"] = "cpu",
        accelerator: Optional[bool] = False,
        nproc: int = 1,
        custom_attr_name_dict: Optional[dict] = None,
    ):

        # model specific parameters
        self.model_map = {
            "gpt3.5": {
                "input_size": 4096,
                "special_token": True,
                "embsize": 512,
            }
        }
        if model_name not in self.model_map:
            raise ValueError(
                f"Model name {model_name} not found in available models: {self.model_map.keys()}"
            )
        list_of_files_to_download = [
            "genept/genept_embeddings/genept_embeddings.json",
        ]

        embeddings_path = Path(
            CACHE_DIR_HELICAL, "genept/genept_embeddings/genept_embeddings.json"
        )

        self.config = {
            "embeddings_path": embeddings_path,
            "model_name": model_name,
            "batch_size": batch_size,
            "emb_layer": emb_layer,
            "emb_mode": emb_mode,
            "device": device,
            "accelerator": accelerator,
            "input_size": self.model_map[model_name]["input_size"],
            "special_token": self.model_map[model_name]["special_token"],
            "embsize": self.model_map[model_name]["embsize"],
            "nproc": nproc,
            "custom_attr_name_dict": custom_attr_name_dict,
            "list_of_files_to_download": list_of_files_to_download,
        }



from omegaconf import OmegaConf
from typing import Literal, List


class TranscriptFormerConfig:
    """
    TranscriptFormerConfig constructor.

    Parameters
    ----------
        model_name: Literal["tf_sapiens", "tf_metazoa", "tf_exemplar"] = "tf_metazoa"
            The name of the model to use.
        batch_size: int = 8
            The number of samples to process in each batch.
        emb_mode: Literal["gene", "cell"] = "cell"
            The mode to use for the embeddings.
        output_keys: List[Literal["gene_llh", "llh"]] = ["gene_llh"]
            The keys to output.
        obs_keys: List[str] = ["all"]
            The keys to include in the output.
        data_files: List[str] = [None]
            Path to input AnnData file(s)
        output_path: str = "./inference_results"
            Directory where results will be saved
        load_checkpoint: str = None
            Path to model weights file (automatically set by inference.py)
        pretrained_embedding: str = None
            Path to pretrained embeddings for out-of-distribution species
        gene_col_name: str = "ensembl_id"
            Column name in AnnData.var containing gene names which will be mapped to ensembl ids. If index is set, .var_names will be used.
        clip_counts: int = 30
            Maximum count value (higher values will be clipped)
        filter_to_vocabs: bool = True
            Whether to filter genes to only those in the vocabulary
        filter_outliers: float = 0.0
            Standard deviation threshold for filtering outlier cells (0.0 = no filtering)
        normalize_to_scale: float = 0
            Scale factor for count normalization (0 = no normalization)
        sort_genes: bool = False
            Whether to sort the genes.
        randomize_genes: bool = False
            Whether to randomize the genes.
        min_expressed_genes: int = 0
            Minimum number of expressed genes required per cell

    """

    def __init__(
        self,
        model_name: Literal["tf_sapiens", "tf_metazoa", "tf_exemplar"] = "tf_sapiens",
        batch_size: int = 8,
        emb_mode: Literal["gene", "cell"] = "cell",
        output_keys: List[Literal["gene_llh", "llh"]] = [
            "llh",
        ],
        obs_keys: List[str] = ["all"],
        data_files: List[str] = [None],
        output_path: str = "./inference_results",
        load_checkpoint: str = None,
        pretrained_embedding: str = None,
        gene_col_name: str = "index",
        clip_counts: int = 30,
        filter_to_vocabs: bool = True,
        filter_outliers: float = 0.0,
        normalize_to_scale: float = 0,
        sort_genes: bool = False,
        randomize_genes: bool = False,
        min_expressed_genes: int = 0,
    ):

        inference_config: dict = {
            "batch_size": batch_size,
            "output_keys": output_keys,
            "obs_keys": obs_keys,
            "data_files": data_files,
            "output_path": output_path,
            "load_checkpoint": load_checkpoint,
            "device": "cuda",
            "pretrained_embedding": pretrained_embedding,
            "emb_mode": emb_mode,
        }

        data_config: dict = {
            "gene_col_name": gene_col_name,
            "clip_counts": clip_counts,
            "filter_to_vocabs": filter_to_vocabs,
            "filter_outliers": filter_outliers,
            "normalize_to_scale": normalize_to_scale,
            "sort_genes": sort_genes,
            "randomize_genes": randomize_genes,
            "min_expressed_genes": min_expressed_genes,
        }

        self.config = OmegaConf.create(
            {
                "model": {
                    "inference_config": inference_config,
                    "data_config": data_config,
                }
            }
        )

        if model_name not in ["tf_sapiens", "tf_metazoa", "tf_exemplar"]:
            raise ValueError(
                f"Model name {model_name} not supported. Only tf_sapiens, tf_metazoa, and tf_exemplar are supported."
            )

        if model_name == "tf_sapiens":
            self.list_of_files_to_download = [
                "transcriptformer/tf_sapiens/config.json",
                "transcriptformer/tf_sapiens/model_weights.pt",
                "transcriptformer/tf_sapiens/vocabs/assay_vocab.json",
                "transcriptformer/tf_sapiens/vocabs/homo_sapiens_gene.h5",
            ]
        elif model_name == "tf_metazoa":
            self.list_of_files_to_download = [
                "transcriptformer/tf_metazoa/config.json",
                "transcriptformer/tf_metazoa/model_weights.pt",
                "transcriptformer/tf_metazoa/vocabs/assay_vocab.json",
                "transcriptformer/tf_metazoa/vocabs/drosophila_melanogaster_gene.h5",
                "transcriptformer/tf_metazoa/vocabs/lytechinus_variegatus_gene.h5",
                "transcriptformer/tf_metazoa/vocabs/plasmodium_falciparum_gene.h5",
                "transcriptformer/tf_metazoa/vocabs/xenopus_laevis_gene.h5",
                "transcriptformer/tf_metazoa/vocabs/caenorhabditis_elegans_gene.h5",
                "transcriptformer/tf_metazoa/vocabs/gallus_gallus_gene.h5",
                "transcriptformer/tf_metazoa/vocabs/mus_musculus_gene.h5",
                "transcriptformer/tf_metazoa/vocabs/saccharomyces_cerevisiae_gene.h5",
                "transcriptformer/tf_metazoa/vocabs/danio_rerio_gene.h5",
                "transcriptformer/tf_metazoa/vocabs/oryctolagus_cuniculus_gene.h5",
                "transcriptformer/tf_metazoa/vocabs/spongilla_lacustris_gene.h5",
                "transcriptformer/tf_metazoa/vocabs/homo_sapiens_gene.h5",
            ]
        elif model_name == "tf_exemplar":
            self.list_of_files_to_download = [
                "transcriptformer/tf_exemplar/config.json",
                "transcriptformer/tf_exemplar/model_weights.pt",
                "transcriptformer/tf_exemplar/vocabs/assay_vocab.json",
                "transcriptformer/tf_exemplar/vocabs/danio_rerio_gene.h5",
                "transcriptformer/tf_exemplar/vocabs/drosophila_melanogaster_gene.h5",
                "transcriptformer/tf_exemplar/vocabs/homo_sapiens_gene.h5",
                "transcriptformer/tf_exemplar/vocabs/mus_musculus_gene.h5",
                "transcriptformer/tf_exemplar/vocabs/caenorhabditis_elegans_gene.h5",
            ]

        self.model_name = model_name



from typing import Optional
from typing import Literal
from helical.constants.paths import CACHE_DIR_HELICAL
from pathlib import Path

SPECIES_GENE_EMBEDDINGS = {
    "ESM2": {
        "human": "Homo_sapiens.GRCh38.gene_symbol_to_embedding_ESM2.pt",
        "mouse": "Mus_musculus.GRCm39.gene_symbol_to_embedding_ESM2.pt",
        "frog": "Xenopus_tropicalis.Xenopus_tropicalis_v9.1.gene_symbol_to_embedding_ESM2.pt",
        "zebrafish": "Danio_rerio.GRCz11.gene_symbol_to_embedding_ESM2.pt",
        "mouse_lemur": "Microcebus_murinus.Mmur_3.0.gene_symbol_to_embedding_ESM2.pt",
        "pig": "Sus_scrofa.Sscrofa11.1.gene_symbol_to_embedding_ESM2.pt",
        "macaca_fascicularis": "Macaca_fascicularis.Macaca_fascicularis_6.0.gene_symbol_to_embedding_ESM2.pt",
        "macaca_mulatta": "Macaca_mulatta.Mmul_10.gene_symbol_to_embedding_ESM2.pt",
    }
}


class UCEConfig:
    """Configuration class to use the Universal Cell-Embedding Model.

    Parameters
    ----------
    model_name : Literal["33l_8ep_1024t_1280", "4layer_model"], optional, default="4layer_model"
        The model name
    batch_size : int, optional, default=24
        The batch size
    species : Literal['human', 'mouse', 'frog', 'zebrafish', "mouse_lemur", "pig", "macaca_fascicularis", "macaca_mulatta"], optional, default="human"
        The species of the data.
    gene_embedding_model: Literal['ESM2'], optional, default="ESM2"
        The gene embedding model to use. For now, we only support ESM2.
    pad_length : int, optional, default=1536
        The padding length
    pad_token_idx : int, optional, default=0
        The padding token index
    chrom_token_left_idx : int, optional, default=1
        The left chrom token index
    chrom_token_right_idx : int, optional, default=2
        The right chrom token index
    cls_token_idx : int, optional, default=3
        The cls token index
    CHROM_TOKEN_OFFSET : int, optional, default=143574
        The chrom token offset
    sample_size : int, optional, default=1024
        The sample size
    CXG : bool, optional, default=True
        Whether to use CXG or not
    output_dim : int, optional, default=1280
        The output dimension
    d_hid : int, optional, default=5120
        The hidden dimension
    token_dim : int, optional, default=5120
        The token dimension
    multi_gpu : bool, optional, default=False
        Whether to use multiple GPUs or not
    device : Literal["cpu", "cuda"], optional, default="cpu"
        The device to use. Either use "cuda" or "cpu".
    accelerator : bool, optional, default=False
        The accelerator configuration. By default same device as model.

    Returns
    -------
    UCEConfig
        The UCE configuration object
    """

    def __init__(
        self,
        model_name: Literal["33l_8ep_1024t_1280", "4layer_model"] = "4layer_model",
        batch_size: int = 24,
        species: Literal[
            "human",
            "mouse",
            "frog",
            "zebrafish",
            "mouse_lemur",
            "pig",
            "macaca_fascicularis",
            "macaca_mulatta",
        ] = "human",
        gene_embedding_model: Literal["ESM2"] = "ESM2",
        pad_length: int = 1536,
        pad_token_idx: int = 0,
        chrom_token_left_idx: int = 1,
        chrom_token_right_idx: int = 2,
        cls_token_idx: int = 3,
        CHROM_TOKEN_OFFSET: int = 143574,
        sample_size: int = 1024,
        CXG: bool = True,
        output_dim: int = 1280,
        d_hid: int = 5120,
        token_dim: int = 5120,
        multi_gpu: bool = False,
        device: Literal["cpu", "cuda"] = "cpu",
        accelerator: Optional[bool] = False,
    ):

        # model specific parameters
        self.model_map = {
            "33l_8ep_1024t_1280": {
                "n_layers": 33,
            },
            "4layer_model": {
                "n_layers": 4,
            },
        }

        if model_name not in self.model_map:
            raise ValueError(
                f"Model name {model_name} not found in available models: {self.model_map.keys()}."
            )

        list_of_files_to_download = [
            "uce/all_tokens.torch",
            f"uce/{model_name}.torch",
            "uce/species_chrom.csv",
            "uce/species_offsets.pkl",
            f"uce/protein_embeddings/{SPECIES_GENE_EMBEDDINGS[gene_embedding_model][species]}",
        ]

        model_path = Path(CACHE_DIR_HELICAL, "uce", f"{model_name}.torch")

        self.config = {
            "model_name": model_name,
            "model_path": model_path,
            "list_of_files_to_download": list_of_files_to_download,
            "batch_size": batch_size,
            "species": species,
            "gene_embedding_model": gene_embedding_model,
            "pad_length": pad_length,
            "pad_token_idx": pad_token_idx,
            "chrom_token_left_idx": chrom_token_left_idx,
            "chrom_token_right_idx": chrom_token_right_idx,
            "cls_token_idx": cls_token_idx,
            "CHROM_TOKEN_OFFSET": CHROM_TOKEN_OFFSET,
            "sample_size": sample_size,
            "CXG": CXG,
            "n_layers": self.model_map[model_name]["n_layers"],
            "output_dim": output_dim,
            "d_hid": d_hid,
            "token_file_path": model_path.parent / "all_tokens.torch",
            "token_dim": token_dim,
            "multi_gpu": multi_gpu,
            "device": device,
            "accelerator": accelerator,
            "embsize": output_dim,
        }



from pathlib import Path
from helical.constants.paths import CACHE_DIR_HELICAL
from typing import Literal


class HelixmRNAConfig:
    """HelixmRNAConfig class to store the configuration of the Helix-mRNA model.

    Parameters
    ----------
    batch_size : int, optional, default=10
        The batch size
    device : Literal["cpu", "cuda"], optional, default="cpu"
        The device to use. Either use "cuda" or "cpu".
    max_length : int, optional, default=12288
        The maximum length of the input sequence.
    nproc: int, optional, default=1
        Number of processes to use for data processing.
    """

    def __init__(
        self,
        batch_size: int = 10,
        device: Literal["cpu", "cuda"] = "cpu",
        max_length: int = 12288,
        nproc: int = 1,
    ):

        model_name: Literal["helical-ai/Helix-mRNA"] = "helical-ai/Helix-mRNA"

        self.config = {
            "model_name": model_name,
            "input_size": max_length,
            "batch_size": batch_size,
            "device": device,
            "nproc": nproc,
        }



from helical.models.caduceus import Caduceus, CaduceusConfig
import hydra
from omegaconf import DictConfig
from pandas import DataFrame


@hydra.main(version_base=None, config_path="configs", config_name="caduceus_config")
def run(cfg: DictConfig):
    caduceus_config = CaduceusConfig(**cfg)
    caduceus = Caduceus(configurer=caduceus_config)

    data = DataFrame({"Sequence": ["ACTG" * int(1024 / 4), "TGCA" * int(1024 / 4)]})
    processed_data = caduceus.process_data(data)

    embeddings = caduceus.get_embeddings(processed_data)
    print(embeddings.shape)


if __name__ == "__main__":
    run()



from helical.models.geneformer import Geneformer, GeneformerConfig
import hydra
from omegaconf import DictConfig
import anndata as ad
from datasets import load_dataset
from helical.utils import get_anndata_from_hf_dataset


@hydra.main(version_base=None, config_path="configs", config_name="geneformer_config")
def run(cfg: DictConfig):
    geneformer_config = GeneformerConfig(**cfg)
    geneformer = Geneformer(configurer=geneformer_config)

    num = geneformer.model.num_parameters(only_trainable=False)
    print(f"Number of parameters: {num:_}".replace("_", " "))

    # either load via huggingface
    # hf_dataset = load_dataset(
    #     "helical-ai/yolksac_human",
    #     split="train[:5%]",
    #     trust_remote_code=True,
    #     download_mode="reuse_cache_if_exists",
    # )
    # ann_data = get_anndata_from_hf_dataset(hf_dataset)

    # or load directly
    ann_data = ad.read_h5ad("./yolksac_human.h5ad")

    dataset = geneformer.process_data(ann_data[:10, :100])
    embeddings = geneformer.get_embeddings(dataset, output_genes=True)

    print(embeddings)
    embeddings, attention_weights = geneformer.get_embeddings(
        dataset, output_attentions=True
    )

    print(embeddings)
    print(attention_weights)


if __name__ == "__main__":
    run()



import json
import os
import torch
from transformers import PreTrainedModel
import re
from .standalone_hyenadna import HyenaDNAModel
import logging

LOGGER = logging.getLogger(__name__)


# helper 1
def inject_substring(orig_str):
    """Hack to handle matching keys between models trained with and without
    gradient checkpointing."""

    # modify for mixer keys
    pattern = r"\.mixer"
    injection = ".mixer.layer"

    modified_string = re.sub(pattern, injection, orig_str)

    # modify for mlp keys
    pattern = r"\.mlp"
    injection = ".mlp.layer"

    modified_string = re.sub(pattern, injection, modified_string)

    return modified_string


# helper 2
def load_weights(scratch_dict, pretrained_dict, checkpointing=False):
    """Loads pretrained (backbone only) weights into the scratch state dict."""

    # loop thru state dict of scratch
    # find the corresponding weights in the loaded model, and set it

    # need to do some state dict "surgery"
    for key, value in scratch_dict.items():
        if "backbone" in key:
            # the state dicts differ by one prefix, '.model', so we add that
            key_loaded = "model." + key
            # breakpoint()
            # need to add an extra ".layer" in key
            if checkpointing:
                key_loaded = inject_substring(key_loaded)
            try:
                scratch_dict[key] = pretrained_dict[key_loaded]
            except:
                raise Exception("key mismatch in the state dicts!")

    # scratch_dict has been updated
    return scratch_dict


class HyenaDNAPreTrainedModel(PreTrainedModel):
    """
    An abstract class to handle weights initialization and a simple interface for downloading and loading pretrained
    models.
    """

    base_model_prefix = "hyenadna"

    def __init__(self):
        pass

    def forward(self, input_ids, **kwargs):
        return self.model(input_ids, **kwargs)

    @classmethod
    def from_pretrained(
        cls,
        config,
        device="cpu",
        use_head=False,
        n_classes=2,
    ):
        scratch_model = HyenaDNAModel(
            **config, use_head=use_head, n_classes=n_classes
        )  # the new model format
        loaded_ckpt = torch.load(
            config["model_path"], map_location=torch.device(device), weights_only=False
        )

        # need to load weights slightly different if using gradient checkpointing
        if config.get("checkpoint_mixer", False):
            checkpointing = (
                config["checkpoint_mixer"] == True or config["checkpoint_mixer"] == True
            )
        else:
            checkpointing = False

        # grab state dict from both and load weights
        state_dict = load_weights(
            scratch_model.state_dict(),
            loaded_ckpt["state_dict"],
            checkpointing=checkpointing,
        )

        # scratch model has now been updated
        scratch_model.load_state_dict(state_dict)
        LOGGER.info("Loaded pretrained weights ok!")
        return scratch_model



from helical.models.mamba2_mrna import Mamba2mRNA, Mamba2mRNAConfig
import pytest
import torch


class TestHelixmRNAModel:
    @pytest.fixture
    def mamba2mRNA(self):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        config = Mamba2mRNAConfig(batch_size=1, device=self.device, max_length=100)
        return Mamba2mRNA(configurer=config)

    @pytest.fixture
    def mock_data(self, mamba2mRNA):
        input_sequences = [
            "AAAA",
            "CCCC",
            "UUUU" * 30,
            "ACGU" * 100,
            "ACGN",
            "ANNU" * 20,
        ]
        tokenized_sequences = mamba2mRNA.process_data(input_sequences)
        return tokenized_sequences

    def test_invalid_sequences(self, mamba2mRNA):
        input_sequences = [
            "AAQA",
            "CCCZ",
            "FJAK" * 30,
            "QWER" * 100,
            "JFHS",
            "OFNW" * 20,
        ]
        with pytest.raises(ValueError, match=r"Invalid RNA sequence:*"):
            mamba2mRNA.process_data(input_sequences)

    def test_helix_mrna_get_embeddings(self, mock_data, mamba2mRNA):
        embeddings = mamba2mRNA.get_embeddings(mock_data)

        assert (
            embeddings is not None
        ), f"Embeddings should not be None for sequence: {mock_data}"



from helical.models.mamba2_mrna import Mamba2mRNAFineTuningModel, Mamba2mRNAConfig
import hydra
from omegaconf import DictConfig


@hydra.main(
    version_base=None,
    config_path="../run_models/configs",
    config_name="mamba2_mrna_config",
)
def run_fine_tuning(cfg: DictConfig):
    input_sequences = ["ACUG" * 20, "AUGC" * 20, "AUGC" * 20, "ACUG" * 20, "AUUG" * 20]
    labels = [0, 2, 2, 0, 1]

    mamba2_mrna_config = Mamba2mRNAConfig(**cfg)
    mamba2_mrna_fine_tune = Mamba2mRNAFineTuningModel(
        mamba2_mrna_config=mamba2_mrna_config,
        fine_tuning_head="classification",
        output_size=3,
    )

    train_dataset = mamba2_mrna_fine_tune.process_data(input_sequences)

    mamba2_mrna_fine_tune.train(train_dataset=train_dataset, train_labels=labels)

    outputs = mamba2_mrna_fine_tune.get_outputs(train_dataset)
    print(outputs)

    # save and load model
    mamba2_mrna_fine_tune.save_model("./mamba2_mrna_fine_tuned_model.pt")
    mamba2_mrna_fine_tune.load_model("./mamba2_mrna_fine_tuned_model.pt")

    outputs = mamba2_mrna_fine_tune.get_outputs(train_dataset)
    print(outputs)

if __name__ == "__main__":
    run_fine_tuning()



from helical.models.hyena_dna import HyenaDNAFineTuningModel, HyenaDNAConfig
import hydra
from omegaconf import DictConfig


@hydra.main(
    version_base=None,
    config_path="../run_models/configs",
    config_name="hyena_dna_config",
)
def run_fine_tuning(cfg: DictConfig):
    input_sequences = ["ACT" * 20, "ATG" * 10, "ATG" * 20, "ACT" * 10, "ATT" * 20]
    labels = [0, 2, 2, 0, 1]

    hyena_dna_config = HyenaDNAConfig(**cfg)
    hyena_dna_fine_tune = HyenaDNAFineTuningModel(
        hyena_config=hyena_dna_config, fine_tuning_head="classification", output_size=3
    )

    train_dataset = hyena_dna_fine_tune.process_data(input_sequences)

    hyena_dna_fine_tune.train(train_dataset=train_dataset, train_labels=labels)

    outputs = hyena_dna_fine_tune.get_outputs(train_dataset)

    print(outputs)

    # save and load model
    hyena_dna_fine_tune.save_model("./hyena_dna_fine_tuned_model.pt")
    hyena_dna_fine_tune.load_model("./hyena_dna_fine_tuned_model.pt")

    outputs = hyena_dna_fine_tune.get_outputs(train_dataset)
    print(outputs)


if __name__ == "__main__":
    run_fine_tuning()



import pytest
import numpy as np
from datasets import Dataset, Features, Value, Sequence
from helical.utils import get_anndata_from_hf_dataset

def create_mock_dataset(gene_names: str):
    data = {
        "raw_counts": [[1, 2, 3, 2], [98], [72, 19]],
        "rows": [[0, 1, 2, 3], [0], [1, 3]],
        "obs1": [10, 20, 30],
        "obs2": [40, 50, 60],
        "size": [4, 4, 4],
    }
    features = Features(
        {
            "raw_counts": Sequence(Value("uint32"), -1, gene_names),
            "rows": Sequence(Value("uint32")),
            "obs1": Value("int64"),
            "obs2": Value("int64"),
            "size": Value("uint32"),
        }
    )
    dataset = Dataset.from_dict(data, features=features)
    return dataset


def test_get_anndata_from_hf_dataset():
    dataset = create_mock_dataset("gene1,gene2,gene3,gene4")
    ann_data = get_anndata_from_hf_dataset(dataset)

    assert ann_data.shape == (3, 4)

    # assert that observation names are correct (ie. no 'rows', 'raw_counts', or 'size')
    assert list(ann_data.obs.columns) == ["obs1", "obs2"]

    # assert that gene names are converted to uppercase
    assert list(ann_data.var_names) == ["GENE1", "GENE2", "GENE3", "GENE4"]
    assert list(ann_data.var["gene_name"]) == ["GENE1", "GENE2", "GENE3", "GENE4"]

    # assert that counts are placed in the correct positions
    assert np.array_equal(
        ann_data.X.toarray(), np.array([[1, 2, 3, 2], [98, 0, 0, 0], [0, 72, 0, 19]])
    )


def test_get_anndata_from_hf_dataset_mismatched_gene_names():
    dataset = create_mock_dataset("gene1,gene2")

    with pytest.raises(ValueError):
        get_anndata_from_hf_dataset(dataset)




from helical.models.uce.gene_embeddings import load_gene_embeddings_adata
from anndata import AnnData
import pandas as pd
import numpy as np
from pathlib import Path
import pytest
from pathlib import Path

CACHE_DIR_HELICAL = Path(Path.home(), ".cache", "helical", "models")


class TestUCEGeneEmbeddings:

    adata = AnnData(
        X=np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
        obs=pd.DataFrame({"species": ["human", "mouse", "rat"]}),
        var=pd.DataFrame({"gene": ["gene1", "gene2", "gene3"]}),
    )
    species = ["human"]
    embedding_model = "ESM2"
    embeddings_path = Path(CACHE_DIR_HELICAL, "uce", "protein_embeddings")

    def test_load_gene_embeddings_adata_filtering_all_genes(self):
        with pytest.raises(ValueError):
            load_gene_embeddings_adata(
                self.adata, self.species, self.embedding_model, self.embeddings_path
            )

    def test_load_gene_embeddings_adata_filtering_no_genes(self):
        self.adata.var_names = ["hoxa6", "cav2", "txk"]
        anndata, mapping_dict = load_gene_embeddings_adata(
            self.adata, self.species, self.embedding_model, self.embeddings_path
        )
        assert (anndata.var_names == ["hoxa6", "cav2", "txk"]).all()
        assert (anndata.obs == self.adata.obs).all().all()
        assert (anndata.X == self.adata.X).all()
        assert len(mapping_dict["human"]) == 19790

    def test_load_gene_embeddings_adata_filtering_some_genes(self):
        self.adata.var_names = ["hoxa6", "cav2", "1"]
        anndata, mapping_dict = load_gene_embeddings_adata(
            self.adata, self.species, self.embedding_model, self.embeddings_path
        )
        assert (anndata.var_names == ["hoxa6", "cav2"]).all()
        assert (anndata.obs == self.adata.obs).all().all()
        assert (anndata.X == [[1, 2], [4, 5], [7, 8]]).all()
        assert len(mapping_dict["human"]) == 19790



import logging

try:
    from .model import Caduceus
    from .caduceus_config import CaduceusConfig
    from .fine_tuning_model import CaduceusFineTuningModel
except:
    LOGGER = logging.getLogger(__name__)
    LOGGER.error(
        "Caduceus not available: If you want to use this model, ensure you have a CUDA GPU and have installed the optional helical[mamba-ssm] dependencies."
    )



import pytest
from helical.models.geneformer import (
    GeneformerConfig,
    Geneformer,
    GeneformerFineTuningModel,
)
from anndata import AnnData
import torch
import pandas as pd
import numpy as np
import anndata as ad

geneformer1_config = GeneformerConfig(model_name="gf-12L-40M-i2048", batch_size=1)
geneformer1 = Geneformer(geneformer1_config)
ann_data = ad.read_h5ad("./yolksac_human.h5ad")
dataset = geneformer1.process_data(ann_data[:1, :100])
embeddings_geneformer_v1 = geneformer1.get_embeddings(dataset, output_genes=False)


class TestGeneformer:
    @pytest.fixture
    def mock_data(self):
        data = AnnData()
        data.var["gene_symbols"] = ["HES4", "PLEKHN1", "SAMD11"]
        data.obs["cell_type"] = ["CD4 T cells"]
        data.X = [[1, 2, 5]]
        return data

    @pytest.fixture
    def mock_embeddings_v1(self, mocker):
        embs = mocker.Mock()
        embs.hidden_states = [
            torch.tensor(
                [
                    [
                        [5.0, 5.0, 5.0, 5.0, 5.0],
                        [1.0, 2.0, 3.0, 2.0, 1.0],
                        [6.0, 6.0, 6.0, 6.0, 6.0],
                    ]
                ]
            )
        ] * 12
        return embs

    @pytest.fixture
    def mock_embeddings_v2(self, mocker):
        embs = mocker.Mock()
        embs.hidden_states = torch.tensor(
            [
                [
                    [6.0, 5.0, 7.0, 5.0, 5.0],
                    [5.0, 5.0, 5.0, 5.0, 5.0],
                    [1.0, 2.0, 3.0, 2.0, 1.0],
                    [6.0, 6.0, 6.0, 6.0, 6.0],
                    [6.0, 6.0, 1.0, 6.0, 2.0],
                ]
            ]
        ).repeat(12, 1, 1, 1)
        return embs

    @pytest.fixture(params=["gf-12L-40M-i2048", "gf-12L-38M-i4096"])
    def geneformer(self, request):
        config = GeneformerConfig(model_name=request.param, batch_size=5)
        geneformer = Geneformer(config)
        return geneformer

    def test_process_data_mapping_to_ensemble_ids(self, geneformer, mock_data):
        # geneformer modifies the anndata in place and maps the gene names to ensembl id
        geneformer.process_data(mock_data, gene_names="gene_symbols")
        assert mock_data.var["ensembl_id"][0] == "ENSG00000188290"
        # is the same as the above line but more verbose (linking the gene symbol to the ensembl id)
        assert (
            mock_data.var[mock_data.var["gene_symbols"] == "SAMD11"][
                "ensembl_id"
            ].values[0]
            == "ENSG00000187634"
        )
        assert (
            mock_data.var[mock_data.var["gene_symbols"] == "PLEKHN1"][
                "ensembl_id"
            ].values[0]
            == "ENSG00000187583"
        )
        assert (
            mock_data.var[mock_data.var["gene_symbols"] == "HES4"]["ensembl_id"].values[
                0
            ]
            == "ENSG00000188290"
        )

    def test_process_data_mapping_to_ensemble_ids_resulting_in_0_genes(
        self, geneformer, mock_data
    ):
        # provide a gene that does not exist in the ensembl database
        mock_data.var["gene_symbols"] = ["1", "2", "3"]
        with pytest.raises(ValueError):
            geneformer.process_data(mock_data, gene_names="gene_symbols")

    @pytest.mark.parametrize(
        "invalid_model_names", ["gf-12L-35M-i2048", "gf-34L-30M-i5000"]
    )
    def test_pass_invalid_model_name(self, invalid_model_names):
        with pytest.raises(ValueError):
            GeneformerConfig(model_name=invalid_model_names)

    def test_ensure_data_validity_raising_error_with_missing_ensembl_id_column(
        self, geneformer, mock_data
    ):
        geneformer.process_data(mock_data, gene_names="gene_symbols")
        del mock_data.var["ensembl_id"]
        with pytest.raises(KeyError):
            geneformer.ensure_rna_data_validity(mock_data, "ensembl_id")

    @pytest.mark.parametrize(
        "gene_symbols, raises_error",
        [
            (["ENSGSAMD11", "ENSGPLEKHN1", "ENSGHES4"], True),  # humans
            (
                ["ENSMUSG00000021033", "ENSMUSG00000021033", "ENSMUSG00000021033"],
                True,
            ),  # mice
            (["SAMD11", "None", "HES4"], True),
            (["SAMD11", "PLEKHN1", "HES4"], False),
        ],
    )
    def test_ensembl_data_is_caught(
        self, geneformer, mock_data, gene_symbols, raises_error
    ):
        mock_data.var["gene_symbols"] = gene_symbols
        if raises_error:
            with pytest.raises(ValueError):
                geneformer.process_data(mock_data, "gene_symbols")
        else:
            geneformer.process_data(mock_data, "gene_symbols")

    def test_cls_mode_with_v1_model_config(self, geneformer):
        if geneformer.config["special_token"]:
            pytest.skip(
                "This test is only for v1 models and should thus be only executed once."
            )
        with pytest.raises(ValueError):
            GeneformerConfig(model_name="gf-12L-40M-i2048", emb_mode="cls")

    @pytest.mark.parametrize("emb_mode", ["cell", "gene"])
    def test_get_embeddings_of_different_modes_v1(
        self, emb_mode, mock_data, mock_embeddings_v1, mocker
    ):
        config = GeneformerConfig(
            model_name="gf-12L-40M-i2048", batch_size=5, emb_mode=emb_mode
        )
        geneformer = Geneformer(config)
        mocker.patch.object(
            geneformer.model, "forward", return_value=mock_embeddings_v1
        )

        dataset = geneformer.process_data(mock_data, gene_names="gene_symbols")
        embeddings = geneformer.get_embeddings(dataset)
        if emb_mode == "gene":
            data_list = pd.Series(
                {
                    "ENSG00000187583": np.array([1.0, 2.0, 3.0, 2.0, 1.0]),
                    "ENSG00000187634": np.array([5.0, 5.0, 5.0, 5.0, 5.0]),
                    "ENSG00000188290": np.array([6.0, 6.0, 6.0, 6.0, 6.0]),
                }
            )
            for key in data_list.index:
                assert np.all(np.equal(embeddings[0][key], data_list[key]))

        if emb_mode == "cell":
            expected = np.array([[4, 4.333333, 4.666667, 4.333333, 4]])
            np.testing.assert_allclose(embeddings, expected, rtol=1e-4, atol=1e-4)

    def test_get_embeddings_with_output_genes(
        self, mock_data, mock_embeddings_v1, mocker
    ):
        config = GeneformerConfig(
            model_name="gf-12L-40M-i2048", batch_size=5, emb_mode="cell"
        )
        geneformer = Geneformer(config)
        mocker.patch.object(
            geneformer.model, "forward", return_value=mock_embeddings_v1
        )

        dataset = geneformer.process_data(mock_data, gene_names="gene_symbols")
        embeddings, genes = geneformer.get_embeddings(dataset, output_genes=True)

        expected = np.array([[4, 4.333333, 4.666667, 4.333333, 4]])
        np.testing.assert_allclose(embeddings, expected, rtol=1e-4, atol=1e-4)
        for gene_list in genes:
            assert len(gene_list) == 3
            assert "ENSG00000187583" in gene_list
            assert "ENSG00000187634" in gene_list
            assert "ENSG00000188290" in gene_list

    @pytest.mark.parametrize("emb_mode", ["cell", "gene", "cls"])
    def test_get_embeddings_of_different_modes_v2(
        self, emb_mode, mock_data, mock_embeddings_v2, mocker
    ):
        config = GeneformerConfig(
            model_name="gf-12L-38M-i4096", batch_size=5, emb_mode=emb_mode
        )
        geneformer = Geneformer(config)
        mocker.patch.object(
            geneformer.model, "forward", return_value=mock_embeddings_v2
        )

        dataset = geneformer.process_data(mock_data, gene_names="gene_symbols")
        embeddings = geneformer.get_embeddings(dataset)
        if emb_mode == "gene":
            data_list = pd.Series(
                {
                    "ENSG00000187583": np.array([1.0, 2.0, 3.0, 2.0, 1.0]),
                    "ENSG00000187634": np.array([5.0, 5.0, 5.0, 5.0, 5.0]),
                    "ENSG00000188290": np.array([6.0, 6.0, 6.0, 6.0, 6.0]),
                }
            )
            for key in data_list.index:
                assert np.all(np.equal(embeddings[0][key], data_list[key]))
        print("Mock Dataset:", mock_embeddings_v2)
        print("Embeddings: ", embeddings)

        if emb_mode == "cls":
            assert (embeddings == np.array([6.0, 5.0, 7.0, 5.0, 5.0])).all()
        if emb_mode == "cell":
            expected = np.array([[4, 4.333333, 4.666667, 4.333333, 4]])
            np.testing.assert_allclose(embeddings, expected, rtol=1e-4, atol=1e-4)

    @pytest.mark.parametrize("emb_mode", ["cell", "gene"])
    def test_fine_tune_classifier_returns_correct_shape(self, emb_mode, mock_data):
        fine_tuned_model = GeneformerFineTuningModel(
            GeneformerConfig(emb_mode=emb_mode),
            fine_tuning_head="classification",
            output_size=1,
        )
        tokenized_dataset = fine_tuned_model.process_data(
            mock_data, gene_names="gene_symbols"
        )
        tokenized_dataset = tokenized_dataset.add_column("labels", list([0]))

        fine_tuned_model.train(train_dataset=tokenized_dataset, label="labels")

        outputs = fine_tuned_model.get_outputs(tokenized_dataset)
        assert outputs.shape == (len(mock_data), 1)

    def test_fine_tune_classifier_cls_returns_correct_shape(self, mock_data):
        fine_tuned_model = GeneformerFineTuningModel(
            GeneformerConfig(model_name="gf-12L-38M-i4096", emb_mode="cls"),
            fine_tuning_head="classification",
            output_size=1,
        )
        tokenized_dataset = fine_tuned_model.process_data(
            mock_data, gene_names="gene_symbols"
        )
        tokenized_dataset = tokenized_dataset.add_column("labels", [0])

        fine_tuned_model.train(train_dataset=tokenized_dataset, label="labels")

        outputs = fine_tuned_model.get_outputs(tokenized_dataset)
        assert outputs.shape == (len(mock_data), 1)

    @pytest.mark.parametrize(
        "model_name,emb_layer,expected_error",
        [
            ("gf-6L-10M-i2048", -1, "No Error"),
            ("gf-6L-10M-i2048", 7, "Error"),
            ("gf-12L-40M-i2048", 6, "No Error"),
            ("gf-20L-151M-i4096", 23, "Error"),
        ],
    )
    def test_embedding_layer_error(self, model_name, emb_layer, expected_error):
        config = GeneformerConfig(
            model_name=model_name, batch_size=5, emb_layer=emb_layer
        )

        if expected_error == "Error":
            with pytest.raises(ValueError):
                Geneformer(config)
        else:
            try:
                Geneformer(config)
            except Exception as error:
                pytest.fail(f"Unexpected error: {error}")

    @pytest.mark.parametrize(
        "model_name,emb_layer",
        [
            ("gf-6L-10M-i2048", -1),
            ("gf-6L-10M-i2048", 3),
            ("gf-12L-40M-i2048", 6),
            ("gf-20L-151M-i4096", -1),
        ],
    )
    def test_layer_to_quant(self, model_name, emb_layer):
        config = GeneformerConfig(
            model_name=model_name, batch_size=5, emb_layer=emb_layer
        )
        geneformer = Geneformer(config)

        assert geneformer.layer_to_quant == emb_layer

    @pytest.mark.parametrize(
        "old_model_name,new_model_name",
        [
            ("gf-6L-30M-i2048", "gf-6L-10M-i2048"),
            ("gf-12L-30M-i2048", "gf-12L-40M-i2048"),
            ("gf-12L-95M-i4096", "gf-12L-38M-i4096"),
            ("gf-12L-95M-i4096-CLcancer", "gf-12L-38M-i4096-CLcancer"),
            ("gf-20L-95M-i4096", "gf-20L-151M-i4096"),
        ],
    )
    def test_model_name_mapping(self, old_model_name, new_model_name):
        config = GeneformerConfig(model_name=old_model_name)

        assert config.config["model_name"] == new_model_name

    def test_embedding_genformer_v1_output(self):
        expected = np.loadtxt(
            "ci/tests/data/geneformer_sample_embeddings.txt", delimiter=","
        )
        np.testing.assert_allclose(
            embeddings_geneformer_v1.squeeze(), expected, rtol=1e-5, atol=1e-5
        )



from helical.models.scgpt import scGPT, scGPTConfig
import hydra
from omegaconf import DictConfig
import anndata as ad
from datasets import load_dataset
from helical.utils import get_anndata_from_hf_dataset


@hydra.main(version_base=None, config_path="configs", config_name="scgpt_config")
def run(cfg: DictConfig):
    scgpt_config = scGPTConfig(**cfg)
    scgpt = scGPT(configurer=scgpt_config)

    # either load via huggingface
    # hf_dataset = load_dataset(
    #     "helical-ai/yolksac_human",
    #     split="train[:5%]",
    #     trust_remote_code=True,
    #     download_mode="reuse_cache_if_exists",
    # )
    # ann_data = get_anndata_from_hf_dataset(hf_dataset)

    # or load directly
    ann_data = ad.read_h5ad("./yolksac_human.h5ad")

    data = scgpt.process_data(ann_data[:10])
    embeddings, input_genes = scgpt.get_embeddings(data, output_genes=True)
    print(embeddings, input_genes)
    embeddings, attn_weights = scgpt.get_embeddings(data, output_attentions=True)

    print(embeddings)
    print(attn_weights.shape)


if __name__ == "__main__":
    run()



import torch


class UCECollator(object):
    def __init__(self, config):
        self.pad_length = config["pad_length"]

    def __call__(self, batch):
        batch_size = len(batch)
        batch_sentences = torch.zeros((batch_size, self.pad_length))
        mask = torch.zeros((batch_size, self.pad_length))
        cell_sentences = torch.zeros((batch_size, self.pad_length))

        idxs = torch.zeros(batch_size)

        i = 0
        max_len = 0
        for bs, msk, idx, seq_len, cs in batch:
            batch_sentences[i, :] = bs
            cell_sentences[i, :] = cs
            max_len = max(max_len, seq_len)
            mask[i, :] = msk
            idxs[i] = idx

            i += 1

        return batch_sentences[:, :max_len], mask[:, :max_len], idxs, cell_sentences



"""
Unit tests for Cell2Sen model using pytest.
"""
import pytest
import numpy as np
import anndata as ad
from datasets import Dataset
import torch
from helical.models.c2s.model import Cell2Sen
from helical.models.c2s.config import Cell2SenConfig

@pytest.fixture(scope="module")
def sample_anndata():
    """Load and prepare a small sample of the yolksac dataset for testing."""
    anndata = ad.read_h5ad("yolksac_human.h5ad")
    # Use a small subset for faster testing
    anndata = anndata[:3, :10].copy()
    return anndata


@pytest.fixture(scope="module")
def cell2sen_model():
    """Initialize Cell2Sen model once for all tests."""
    config = Cell2SenConfig(batch_size=8)
    return Cell2Sen(configurer=config)


@pytest.fixture(scope="module")
def processed_dataset_basic(cell2sen_model, sample_anndata):
    """Pre-processed dataset for basic tests."""
    return cell2sen_model.process_data(sample_anndata)


@pytest.fixture(scope="module")
def processed_dataset_with_perturbation(cell2sen_model, sample_anndata):
    """Pre-processed dataset with perturbation column."""
    sample_anndata_copy = sample_anndata.copy()
    sample_anndata_copy.obs['perturbation'] = 'IFNg'
    
    # Temporarily modify the shared model's perturbation_column
    original_pert_col = cell2sen_model.perturbation_column
    cell2sen_model.perturbation_column = 'perturbation'
    
    dataset = cell2sen_model.process_data(sample_anndata_copy)
    cell2sen_model.perturbation_column = original_pert_col
    
    return dataset


@pytest.fixture(scope="module")
def processed_dataset_linear_map(cell2sen_model, sample_anndata):
    """Pre-processed dataset with linear expression map enabled."""
    # Temporarily modify the shared model's return_fit
    original_return_fit = cell2sen_model.return_fit
    cell2sen_model.return_fit = True
    
    dataset = cell2sen_model.process_data(sample_anndata)
    cell2sen_model.return_fit = original_return_fit
    
    return dataset


class TestInit:
    """Test Cell2Sen initialization."""
    
    def test_init_default(self):
        """Test model initialization with default parameters."""
        model = Cell2Sen()
        assert model is not None
        assert hasattr(model, 'model')
        assert hasattr(model, 'tokenizer')
        assert hasattr(model, 'device')
        assert model.model is not None
        assert model.tokenizer is not None
    
    def test_init_with_config(self):
        """Test model initialization with custom config."""
        config = Cell2SenConfig(batch_size=16, organism="Homo sapiens")
        model = Cell2Sen(configurer=config)
        assert model.batch_size == 16
        assert model.organism == "Homo sapiens"
    
    def test_model_on_correct_device(self, cell2sen_model):
        """Test that model is on the correct device."""
        expected_device = "cuda" if torch.cuda.is_available() else "cpu"
        assert cell2sen_model.device == expected_device
    
    def test_model_in_eval_mode(self, cell2sen_model):
        """Test that model is in evaluation mode."""
        assert not cell2sen_model.model.training
    
    def test_tokenizer_loaded(self, cell2sen_model):
        """Test that tokenizer is properly loaded."""
        test_text = "Test sentence"
        tokens = cell2sen_model.tokenizer(test_text, return_tensors="pt")
        assert tokens is not None
        assert 'input_ids' in tokens


class TestProcessData:
    """Test process_data method."""
    
    def test_process_data_basic_properties(self, cell2sen_model, sample_anndata):
        """Test basic data processing - checks all properties in one go."""
        dataset = cell2sen_model.process_data(sample_anndata)
        
        # Check dataset type and structure
        assert isinstance(dataset, Dataset)
        assert len(dataset) == sample_anndata.n_obs
        
        # Check all required columns exist
        required_columns = ['cell_sentence', 'fit_parameters', 'organism', 'perturbations']
        assert all(col in dataset.column_names for col in required_columns)
        
        # Check cell sentences format
        assert all(isinstance(s, str) for s in dataset['cell_sentence'])
        assert len(dataset['cell_sentence'][0].split()) > 0
        
        # Check organism format
        assert all(isinstance(org, str) for org in dataset['organism'])
        assert all(org == dataset['organism'][0] for org in dataset['organism'])
        
        # Check perturbations are None when not provided
        assert all(p is None for p in dataset['perturbations'])
    
    def test_process_data_with_perturbation(self, cell2sen_model, sample_anndata):
        """Test data processing with perturbation column."""
        sample_anndata_copy = sample_anndata.copy()
        sample_anndata_copy.obs['perturbation'] = 'IFNg'
        
        # Temporarily modify the shared model
        original_pert_col = cell2sen_model.perturbation_column
        cell2sen_model.perturbation_column = 'perturbation'
        
        dataset = cell2sen_model.process_data(sample_anndata_copy)
        assert len(dataset['perturbations']) == len(dataset)
        assert all(p == 'IFNg' for p in dataset['perturbations'])
        cell2sen_model.perturbation_column = original_pert_col
    
    def test_process_data_with_perturbation_list(self, cell2sen_model, sample_anndata):
        """Test data processing with list of perturbations."""
        sample_anndata_copy = sample_anndata.copy()
        perturbations = ['IFNg', 'TNF', 'IL-1'] * (sample_anndata.n_obs // 3 + 1)
        perturbations = perturbations[:sample_anndata.n_obs]
        sample_anndata_copy.obs['perturbation'] = perturbations
        
        # Temporarily modify the shared model
        original_pert_col = cell2sen_model.perturbation_column
        cell2sen_model.perturbation_column = 'perturbation'
        
        dataset = cell2sen_model.process_data(sample_anndata_copy)
        assert len(dataset['perturbations']) == len(dataset)
        assert dataset['perturbations'][0] == 'IFNg'
        cell2sen_model.perturbation_column = original_pert_col
    
    def test_process_data_organism_provided(self, cell2sen_model, sample_anndata):
        """Test that organism is properly stored when provided."""
        organism = "Homo sapiens"
        
        # Temporarily modify the shared model
        original_organism = cell2sen_model.organism
        cell2sen_model.organism = organism
        
        dataset = cell2sen_model.process_data(sample_anndata)
        assert 'organism' in dataset.column_names
        assert len(dataset['organism']) == len(dataset)
        assert all(isinstance(org, str) for org in dataset['organism'])
        assert all(org == organism for org in dataset['organism'])
        cell2sen_model.organism = original_organism
    
    def test_process_data_fit_parameters(self, processed_dataset_linear_map):
        """Test that fit parameters are calculated correctly."""
        assert len(processed_dataset_linear_map['fit_parameters']) == len(processed_dataset_linear_map)
        first_params = processed_dataset_linear_map['fit_parameters'][0]
        if first_params is not None:
            assert 'slope' in first_params
            assert 'intercept' in first_params
            assert 'r_squared' in first_params
    
    def test_process_data_fit_parameters_disabled(self, processed_dataset_basic):
        """Test that fit parameters are None when disabled."""
        assert all(fp is None for fp in processed_dataset_basic['fit_parameters'])


class TestGetEmbeddings:
    """Test get_embeddings method."""
    
    def test_get_embeddings_comprehensive(self, cell2sen_model, processed_dataset_basic):
        """Test embeddings comprehensively - shape, type, dimensions."""
        embeddings = cell2sen_model.get_embeddings(processed_dataset_basic)
        
        # Check type and shape
        assert isinstance(embeddings, np.ndarray)
        assert embeddings.ndim == 2
        assert embeddings.shape[0] == len(processed_dataset_basic)
        assert embeddings.shape[1] == 2304  # Gemma 2 2B hidden size
        assert embeddings.dtype == np.float32
        
    
    def test_get_embeddings_with_different_batch_sizes(self, cell2sen_model, processed_dataset_basic):
        """Test embeddings have consistent shape across batch sizes."""
        # Use the shared model but change batch size dynamically
        original_batch_size = cell2sen_model.batch_size
        
        # Test with batch size 2
        cell2sen_model.batch_size = 2
        embeddings_batch2 = cell2sen_model.get_embeddings(processed_dataset_basic)
        
        # Test with batch size 3
        cell2sen_model.batch_size = 3
        embeddings_batch5 = cell2sen_model.get_embeddings(processed_dataset_basic)
        
        # Only check shape - values may differ due to padding differences
        assert embeddings_batch2.shape == embeddings_batch5.shape
        assert embeddings_batch2.shape[0] == len(processed_dataset_basic)
        # Restore original batch size
        cell2sen_model.batch_size = original_batch_size

    
    def test_get_embeddings_empty_dataset(self, cell2sen_model, sample_anndata):
        """Test embeddings with empty dataset."""
        empty_dataset = cell2sen_model.process_data(sample_anndata[:0])
        with pytest.raises((ValueError, IndexError, AssertionError)):
            cell2sen_model.get_embeddings(empty_dataset)

    def test_get_embeddings_attention_shapes(self, cell2sen_model, processed_dataset_basic):
        """Test that attention maps have correct shapes when output_attentions=True."""
        embeddings, attentions = cell2sen_model.get_embeddings(
            processed_dataset_basic, 
            output_attentions=True
        )
        
        # Check embeddings shape
        assert isinstance(embeddings, np.ndarray)
        assert embeddings.shape[0] == len(processed_dataset_basic)
        
        # Check attentions is a tuple (one element per layer)
        assert isinstance(attentions, tuple)
        assert len(attentions) > 0  # Should have at least one layer
        
        # Check each layer's attention map shape
        num_cells = len(processed_dataset_basic)
        for layer_idx, attn_layer in enumerate(attentions):
            assert isinstance(attn_layer, np.ndarray)
            assert attn_layer.ndim == 4, f"Layer {layer_idx} attention should be 4D"
            assert attn_layer.shape[0] == num_cells, f"Layer {layer_idx} batch dimension should match number of cells"
            assert attn_layer.shape[1] > 0, f"Layer {layer_idx} should have at least one attention head"
            assert attn_layer.shape[2] == attn_layer.shape[3], f"Layer {layer_idx} attention should be square (seq_len x seq_len)"
            assert attn_layer.shape[2] > 0, f"Layer {layer_idx} sequence length should be positive"

class TestGetPerturbations:
    """Test get_perturbations method."""
    
    def test_get_perturbations_comprehensive(self, cell2sen_model, processed_dataset_with_perturbation):
        """Test perturbation generation comprehensively."""
        # Temporarily modify batch_size
        original_batch_size = cell2sen_model.batch_size
        cell2sen_model.batch_size = 5
        
        dataset, perturbed_sentences = cell2sen_model.get_perturbations(processed_dataset_with_perturbation)
        assert isinstance(perturbed_sentences, list)
        assert len(perturbed_sentences) == len(dataset)
        assert all(isinstance(s, str) or s is None for s in perturbed_sentences)
        assert 'perturbed_cell_sentence' in dataset.column_names
        assert len(dataset['perturbed_cell_sentence']) == len(dataset)
        assert all(s is not None for s in perturbed_sentences)
        assert all(isinstance(s, str) for s in perturbed_sentences)
        assert all(len(s) > 0 for s in perturbed_sentences)
        cell2sen_model.batch_size = original_batch_size
    
    def test_get_perturbations_with_perturbations_list(self, cell2sen_model, processed_dataset_basic):
        """Test perturbation generation with provided perturbations_list."""
        perturbations_list = ['IFNg'] * len(processed_dataset_basic)
        dataset, perturbed_sentences = cell2sen_model.get_perturbations(
            processed_dataset_basic,
            perturbations_list=perturbations_list
        )
        
        assert len(perturbed_sentences) == len(processed_dataset_basic)
        assert all(s is not None for s in perturbed_sentences)
        
    
    def test_get_perturbations_with_mixed_none(self, cell2sen_model, sample_anndata):
        """Test perturbation generation with mixed None values."""
        sample_anndata_copy = sample_anndata.copy()
        perturbations = ['IFNg', None, 'TNF', None, 'IL-1'] * (sample_anndata.n_obs // 5 + 1)
        perturbations = perturbations[:sample_anndata.n_obs]
        sample_anndata_copy.obs['perturbation'] = perturbations
        
        # Temporarily modify the shared model
        original_pert_col = cell2sen_model.perturbation_column
        cell2sen_model.perturbation_column = 'perturbation'
        
        dataset = cell2sen_model.process_data(sample_anndata_copy)
        dataset, perturbed_sentences = cell2sen_model.get_perturbations(dataset)
        assert len(perturbed_sentences) == len(dataset)
        assert any(s is None for s in perturbed_sentences)
        assert any(s is not None for s in perturbed_sentences)
        cell2sen_model.perturbation_column = original_pert_col
    
    def test_get_perturbations_all_none_error(self, cell2sen_model, processed_dataset_basic):
        """Test error when all perturbations are None."""
        with pytest.raises(ValueError, match="No valid perturbations"):
            cell2sen_model.get_perturbations(processed_dataset_basic)
    
    def test_get_perturbations_length_mismatch_error(self, cell2sen_model, processed_dataset_basic):
        """Test error when perturbations_list length doesn't match dataset."""
        perturbations_list = ['IFNg'] * 5  # Wrong length
        
        with pytest.raises(ValueError, match="must match dataset length"):
            cell2sen_model.get_perturbations(
                processed_dataset_basic,
                perturbations_list=perturbations_list
            )


from helical.constants.paths import CACHE_DIR_HELICAL
from pathlib import Path

EMBEDDING_PROMPT = """
You are given a list of genes in descending order of expression levels in a {organism} cell. \n
Genes: {cell_sentence} \n
Using this information, predict the cell type. 
"""

PERTURBATION_PROMPT = """
You are given a list of genes in descending order of expression levels in a {organism} control cell. \n
Using this information, predict the perturbed gene profile with the perturbation: {perturbation}.
Control Cell Sentence: {cell_sentence} \n
Perturbed Cell Sentence:
"""

class Cell2SenConfig:
    """
    Configuration class for the Cell2Sen Model.

    Parameters
    ----------
    batch_size: int = 16
        int: Number of samples to process in each batch during model operations. Default is 16.
    
    organism: str = None
        The organism from which the cell data is derived (e.g., 'human', 'mouse').

    perturbation_column: str = None
        Column name in the input data that specifies the perturbation applied to cells.

    max_new_tokens: int = 200
        Maximum number of new tokens that the model can generate for prediction. Default is 200.
        One gene is roughly 4 tokens. 

    return_fit: bool = False
        Whether to return model fit parameters in outputs. Default is False. This fits a linear model (y=mx+c) to the gene rank and expression values in log10-transformed space
        and can be used to map between expression values and gene ranks. The paper shows this is well captured by a linear model. The fit parameters are returned in the `fit_parameters` field.
    
    dtype: str = "bfloat16"
        Data type for the model. Default is "bfloat16". 
    
    model_size: str = "2B"
        Size of the model. Default is "2B".
        Choices are "2B" or "27B".
    
    use_quantization: bool = False
        Whether to use 4-bit quantization. Default is False.

    seed: int = 42
        Random seed for reproducibility. Default is 42.

    """
    def __init__(
        self,
        batch_size: int = 16,
        organism: str = None,
        perturbation_column: str = None,
        max_new_tokens: int = 200,
        return_fit: bool = False,
        dtype: str = "bfloat16", 
        model_size: str = "2B",
        use_quantization: bool = False,
        seed: int = 42,
    ):

        if model_size == "2B":
            model_path = Path(CACHE_DIR_HELICAL, "c2s_model_2B")
            hf_model_path = "vandijklab/C2S-Scale-Gemma-2-2B"
            # list_of_files_to_download = [
            #     "c2s_model/config.json",
            #     "c2s_model/generation_config.json",
            #     "c2s_model/model-00001-of-00002.safetensors",
            #     "c2s_model/model-00002-of-00002.safetensors",
            #     "c2s_model/model.safetensors.index.json",
            #     "c2s_model/special_tokens_map.json",
            #     "c2s_model/tokenizer_config.json",
            #     "c2s_model/tokenizer.json",
            # ]
        elif model_size == "27B":
            model_path = Path(CACHE_DIR_HELICAL, "c2s_model_27B")
            hf_model_path = "vandijklab/C2S-Scale-Gemma-2-27B"
            # list_of_files_to_download = [
            #     "c2s_model_27B/config.json",
            #     "c2s_model_27B/generation_config.json",
            #     "c2s_model_27B/model-00001-of-00012.safetensors",
            #     "c2s_model_27B/model-00002-of-00012.safetensors",
            #     "c2s_model_27B/model-00003-of-00012.safetensors",
            #     "c2s_model_27B/model-00004-of-00012.safetensors",
            #     "c2s_model_27B/model-00005-of-00012.safetensors",
            #     "c2s_model_27B/model-00006-of-00012.safetensors",
            #     "c2s_model_27B/model-00007-of-00012.safetensors",
            #     "c2s_model_27B/model-00008-of-00012.safetensors",
            #     "c2s_model_27B/model-00009-of-00012.safetensors",
            #     "c2s_model_27B/model-00010-of-00012.safetensors",
            #     "c2s_model_27B/model-00011-of-00012.safetensors",
            #     "c2s_model_27B/model-00012-of-00012.safetensors",
            #     "c2s_model_27B/model.safetensors.index.json",
            #     "c2s_model_27B/special_tokens_map.json",
            #     "c2s_model_27B/tokenizer_config.json",
            #     "c2s_model_27B/tokenizer.json",
            # ]
        else:
            raise ValueError(f"Model size {model_size} not supported. Please choose from '2B' or '27B'.")

        self.config = {
            "hf_model_path": hf_model_path,
            # "list_of_files_to_download": list_of_files_to_download,
            "model_path": model_path,
            "batch_size": batch_size,
            "organism": organism,
            "perturbation_column": perturbation_column,
            "max_new_tokens": max_new_tokens,
            "return_fit": return_fit,
            "use_quantization": use_quantization,
            "seed": seed,
            "dtype": dtype,
            "model_size": model_size,
        }


# based on https://github.com/EleutherAI/gpt-neox/blob/main/megatron/tokenizer/tokenizer.py
from abc import ABC, abstractmethod
from typing import List, Union

import numpy as np
import torch


class AbstractTokenizer(ABC):
    """Abstract class for tokenizer."""

    def __init__(self, name):
        self.name = name
        super().__init__()

    @property
    @abstractmethod
    def vocab_size(self):
        pass

    @property
    @abstractmethod
    def vocab(self):
        """Dictionary from vocab text token to id token."""
        pass

    @property
    @abstractmethod
    def inv_vocab(self):
        """Dictionary from vocab id token to text token."""
        pass

    @abstractmethod
    def tokenize(self, text):
        pass

    def detokenize(self, token_ids):
        raise NotImplementedError(
            "detokenizer is not implemented for {} " "tokenizer".format(self.name)
        )

    @property
    def cls(self):
        raise NotImplementedError(
            "CLS is not provided for {} " "tokenizer".format(self.name)
        )

    @property
    def sep(self):
        raise NotImplementedError(
            "SEP is not provided for {} " "tokenizer".format(self.name)
        )

    @property
    def pad(self):
        raise NotImplementedError(
            "PAD is not provided for {} " "tokenizer".format(self.name)
        )

    @property
    def eod(self):
        raise NotImplementedError(
            "EOD is not provided for {} " "tokenizer".format(self.name)
        )

    @property
    def mask(self):
        raise NotImplementedError(
            "MASK is not provided for {} " "tokenizer".format(self.name)
        )


class CharLevelTokenizer(AbstractTokenizer):
    """Character Level Tokenizer"""

    def __init__(self, vocab_size):
        name = "CharLevelTokenizer"
        super().__init__(name)
        self._vocab_size = vocab_size
        self.eod_id = 0
        self.eos_id = 0
        self.pad_id = 1

    def clamp(self, n):
        return max(32, min(n, self.vocab_size))

    @property
    def vocab_size(self):
        return self._vocab_size

    @property
    def vocab(self):
        raise NotImplementedError

    @property
    def inv_vocab(self):
        raise NotImplementedError

    def decode_token(self, token: int):
        return str(chr(self.clamp(token)))

    def tokenize(self, text: str):
        return list(np.fromstring(text, dtype=np.uint8))

    def tokenize_batch(self, text_batch: Union[List[str], str]):
        if isinstance(text_batch, list):
            return [self.tokenize(s) for s in text_batch]
        else:
            return self.tokenize(text_batch)

    def detokenize(self, token_ids):
        return "".join(list(map(self.decode_token, token_ids)))

    def detokenize_batch(self, token_ids: Union[List[str], str]):
        if isinstance(token_ids, list):
            return [self.detokenize(s) for s in token_ids]
        # elif if tensor, convert to list first
        elif isinstance(token_ids, torch.Tensor):
            return [self.detokenize(s) for s in token_ids.tolist()]
        else:
            return self.detokenize(token_ids)

    @property
    def eod(self):
        return self.eod_id

    # duplicate to suppose both names, eos and eod
    @property
    def eos(self):
        return self.eod_id



from typing import Optional, Tuple

import torch
from torch import nn


# The code is modified from https://github.com/wgchang/DSBN/blob/master/model/dsbn.py
class _DomainSpecificBatchNorm(nn.Module):
    _version = 2

    def __init__(
        self,
        num_features: int,
        num_domains: int,
        eps: float = 1e-5,
        momentum: float = 0.1,
        affine: bool = True,
        track_running_stats: bool = True,
    ):
        super(_DomainSpecificBatchNorm, self).__init__()
        self._cur_domain = None
        self.num_domains = num_domains
        self.bns = nn.ModuleList(
            [
                self.bn_handle(num_features, eps, momentum, affine, track_running_stats)
                for _ in range(num_domains)
            ]
        )

    @property
    def bn_handle(self) -> nn.Module:
        raise NotImplementedError

    @property
    def cur_domain(self) -> Optional[int]:
        return self._cur_domain

    @cur_domain.setter
    def cur_domain(self, domain_label: int):
        self._cur_domain = domain_label

    def reset_running_stats(self):
        for bn in self.bns:
            bn.reset_running_stats()

    def reset_parameters(self):
        for bn in self.bns:
            bn.reset_parameters()

    def _check_input_dim(self, input: torch.Tensor):
        raise NotImplementedError

    def forward(self, x: torch.Tensor, domain_label: int) -> torch.Tensor:
        self._check_input_dim(x)
        if domain_label >= self.num_domains:
            raise ValueError(
                f"Domain label {domain_label} exceeds the number of domains {self.num_domains}"
            )
        bn = self.bns[domain_label]
        self.cur_domain = domain_label
        return bn(x)


class DomainSpecificBatchNorm1d(_DomainSpecificBatchNorm):
    @property
    def bn_handle(self) -> nn.Module:
        return nn.BatchNorm1d

    def _check_input_dim(self, input: torch.Tensor):
        if input.dim() > 3:
            raise ValueError(
                "expected at most 3D input (got {}D input)".format(input.dim())
            )


class DomainSpecificBatchNorm2d(_DomainSpecificBatchNorm):
    @property
    def bn_handle(self) -> nn.Module:
        return nn.BatchNorm2d

    def _check_input_dim(self, input: torch.Tensor):
        if input.dim() != 4:
            raise ValueError("expected 4D input (got {}D input)".format(input.dim()))
